# --------------------------------------------------------------------------- #
#   timer.py                                                                  #
#                                                                             #
#   Copyright © 2015-2022, Rajiv Bakulesh Shah, original author.              #
#                                                                             #
#   Licensed under the Apache License, Version 2.0 (the "License");           #
#   you may not use this file except in compliance with the License.          #
#   You may obtain a copy of the License at:                                  #
#       http://www.apache.org/licenses/LICENSE-2.0                            #
#                                                                             #
#   Unless required by applicable law or agreed to in writing, software       #
#   distributed under the License is distributed on an "AS IS" BASIS,         #
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  #
#   See the License for the specific language governing permissions and       #
#   limitations under the License.                                            #
# --------------------------------------------------------------------------- #
'Measure the execution time of small code snippets.'


# TODO: Remove the following import after deferred evaluation of annotations
# because the default.
#   1. https://docs.python.org/3/whatsnew/3.7.html#whatsnew37-pep563
#   2. https://www.python.org/dev/peps/pep-0563/
#   3. https://www.python.org/dev/peps/pep-0649/
from __future__ import annotations

import timeit
from types import TracebackType
from typing import Type
from typing import overload

from typing_extensions import Literal


class ContextTimer:
    '''Measure the execution time of small code snippets.

    Note that ContextTimer measures wall (real-world) time, not CPU time; and
    that .elapsed() returns time in milliseconds.

    You can use ContextTimer stand-alone...

        >>> import time
        >>> timer = ContextTimer()
        >>> timer.start()
        >>> time.sleep(0.1)
        >>> 100 <= timer.elapsed() < 200
        True
        >>> timer.stop()
        >>> time.sleep(0.1)
        >>> 100 <= timer.elapsed() < 200
        True

    ...or as a context manager:

        >>> tests = []
        >>> with ContextTimer() as timer:
        ...     time.sleep(0.1)
        ...     tests.append(100 <= timer.elapsed() < 200)
        >>> time.sleep(0.1)
        >>> tests.append(100 <= timer.elapsed() < 200)
        >>> tests
        [True, True]
    '''

    __slots__ = ('_started', '_stopped')

    def __init__(self) -> None:
        self._started = 0.0
        self._stopped = 0.0

    def __enter__(self) -> ContextTimer:
        self.__start()
        return self

    @overload
    def __exit__(self,
                 exc_type: None,
                 exc_value: None,
                 exc_traceback: None,
                 ) -> Literal[False]:
        raise NotImplementedError

    @overload
    def __exit__(self,
                 exc_type: Type[BaseException],
                 exc_value: BaseException,
                 exc_traceback: TracebackType,
                 ) -> Literal[False]:
        raise NotImplementedError

    def __exit__(self,
                 exc_type: Type[BaseException] | None,
                 exc_value: BaseException | None,
                 exc_traceback: TracebackType | None,
                 ) -> Literal[False]:
        self.__stop()
        return False

    def start(self) -> None:
        if self._stopped:
            raise RuntimeError('timer has already been stopped')
        elif self._started:
            raise RuntimeError('timer has already been started')
        else:
            self._started = timeit.default_timer()

    # Preserve the Open-Closed Principle with name mangling.
    #   https://youtu.be/miGolgp9xq8?t=2086
    #   https://stackoverflow.com/a/38534939
    __start = start

    def stop(self) -> None:
        if self._stopped:
            raise RuntimeError('timer has already been stopped')
        elif self._started:
            self._stopped = timeit.default_timer()
        else:
            raise RuntimeError("timer hasn't yet been started")

    __stop = stop

    def elapsed(self) -> int:
        if self._started:
            stopped_or_current = self._stopped or timeit.default_timer()
            elapsed = stopped_or_current - self._started
            return round(elapsed * 1000)
        else:
            raise RuntimeError("timer hasn't yet been started")



# --------------------------------------------------------------------------- #
#   queue.py                                                                  #
#                                                                             #
#   Copyright © 2015-2022, Rajiv Bakulesh Shah, original author.              #
#                                                                             #
#   Licensed under the Apache License, Version 2.0 (the "License");           #
#   you may not use this file except in compliance with the License.          #
#   You may obtain a copy of the License at:                                  #
#       http://www.apache.org/licenses/LICENSE-2.0                            #
#                                                                             #
#   Unless required by applicable law or agreed to in writing, software       #
#   distributed under the License is distributed on an "AS IS" BASIS,         #
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  #
#   See the License for the specific language governing permissions and       #
#   limitations under the License.                                            #
# --------------------------------------------------------------------------- #


# TODO: When we drop support for Python 3.9, remove the following import.  We
# only need it for X | Y union type annotations as of 2022-01-29.
from __future__ import annotations

import math
import random
import time
from typing import ClassVar
from typing import Tuple
from typing import cast

from redis import WatchError

from .annotations import JSONTypes
from .base import Container
from .exceptions import QueueEmptyError
from .timer import ContextTimer


class RedisSimpleQueue(Container):
    RETRY_DELAY: ClassVar[int] = 200

    def qsize(self) -> int:
        '''Return the size of the queue.  O(1)

        Be aware that there's a potential race condition here where the queue
        changes before you use the result of .qsize().
        '''
        return self.redis.xlen(self.key)  # Available since Redis 5.0.0

    # Preserve the Open-Closed Principle with name mangling.
    #   https://youtu.be/miGolgp9xq8?t=2086
    #   https://stackoverflow.com/a/38534939
    __qsize = qsize

    def empty(self) -> bool:
        '''Return True if the queue is empty; False otherwise.  O(1)

        This method is likely to be removed at some point.  Use `.qsize() == 0`
        as a direct substitute, but be aware that either approach risks a race
        condition where the queue grows before you use the result of .empty() or
        .qsize().
        '''
        return self.__qsize() == 0

    def put(self,
            item: JSONTypes,
            block: bool = True,
            timeout: float | None = None,
            ) -> None:
        '''Put the item on the queue.  O(1)

        The optional block and timeout arguments are ignored, as this method
        never blocks.  They are provided for compatibility with the queue.Queue
        class.
        '''
        encoded_item = self._encode(item)
        self.redis.xadd(self.key, {'item': encoded_item}, id='*')  # Available since Redis 5.0.0

    __put = put

    def put_nowait(self, item: JSONTypes) -> None:
        '''Put an item into the queue without blocking.  O(1)

        This is exactly equivalent to `.put(item)` and is provided for
        compatibility with the queue.Queue class.
        '''
        return self.__put(item, block=False)

    def get(self,
            block: bool = True,
            timeout: float | None = None,
            ) -> JSONTypes:
        '''Remove and return an item from the queue.  O(1)

        If optional args block is True and timeout is None (the default), block
        if necessary until an item is available.  If timeout is a non-negative
        number, block at most timeout seconds and raise the QueueEmptyError
        exception if no item becomes available within that time.  Otherwise
        (block is False), return an item if one is immediately available, else
        raise the QueueEmptyError exception (timeout is ignored in this case).
        '''
        redis_block = (timeout or 0.0) if block else 0.0
        redis_block = math.floor(redis_block)
        with ContextTimer() as timer:
            while True:
                try:
                    item = self.__remove_and_return(redis_block)
                    return item
                except (WatchError, IndexError):
                    if not block or timer.elapsed() / 1000 >= (timeout or 0):
                        raise QueueEmptyError(redis=self.redis, key=self.key)
                    delay = random.uniform(0, self.RETRY_DELAY/1000)  # nosec
                    time.sleep(delay)

    __get = get

    def __remove_and_return(self, redis_block: int) -> JSONTypes:
        with self._watch() as pipeline:
            # XXX: The following line raises WatchError after the socket timeout
            # if the RedisQueue is empty and we're not blocking.  This feels
            # like a bug in redis-py?
            returned_value = pipeline.xread(  # Available since Redis 5.0.0
                {self.key: 0},
                count=1,
                block=redis_block,
            )
            # The following line raises IndexError if the RedisQueue is empty
            # and we're blocking.
            id_, dict_ = cast(Tuple[bytes, dict], returned_value[0][1][0])
            pipeline.multi()  # Available since Redis 1.2.0
            pipeline.xdel(self.key, id_)  # Available since Redis 5.0.0
        encoded_item = dict_[b'item']
        item = self._decode(encoded_item)
        return item

    def get_nowait(self) -> JSONTypes:
        '''Remove and return an item from the queue without blocking.  O(1)

        Get an item if one is immediately available.  Otherwise raise the
        QueueEmptyError exception.

        This is exactly equivalent to `.get(block=False)` and is provided for
        compatibility with the queue.Queue class.
        '''
        return self.__get(block=False)



# --------------------------------------------------------------------------- #
#   dict.py                                                                   #
#                                                                             #
#   Copyright © 2015-2022, Rajiv Bakulesh Shah, original author.              #
#                                                                             #
#   Licensed under the Apache License, Version 2.0 (the "License");           #
#   you may not use this file except in compliance with the License.          #
#   You may obtain a copy of the License at:                                  #
#       http://www.apache.org/licenses/LICENSE-2.0                            #
#                                                                             #
#   Unless required by applicable law or agreed to in writing, software       #
#   distributed under the License is distributed on an "AS IS" BASIS,         #
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  #
#   See the License for the specific language governing permissions and       #
#   limitations under the License.                                            #
# --------------------------------------------------------------------------- #


# TODO: When we drop support for Python 3.9, remove the following import.  We
# only need it for X | Y union type annotations as of 2022-01-29.
from __future__ import annotations

import collections.abc
import itertools
import warnings
from typing import Any
from typing import Dict
from typing import Generator
from typing import Iterable
from typing import Mapping
from typing import Tuple
from typing import Union
from typing import cast

from redis import Redis
from redis.client import Pipeline

from .annotations import JSONTypes
from .base import Container
from .base import Iterable_
from .exceptions import InefficientAccessWarning
from .exceptions import KeyExistsError


InitMap = Mapping[JSONTypes, JSONTypes]
InitItem = Tuple[JSONTypes, JSONTypes]
InitIter = Iterable[InitItem]
InitArg = Union[InitMap, InitIter]


class RedisDict(Container, Iterable_, collections.abc.MutableMapping):
    'Redis-backed container compatible with Python dicts.'

    def __init__(self,
                 arg: InitArg = tuple(),
                 *,
                 redis: Redis | None = None,
                 key: str = '',
                 **kwargs: JSONTypes,
                 ) -> None:
        'Initialize the RedisDict.  O(n)'
        super().__init__(redis=redis, key=key)
        if arg or kwargs:
            with self._watch(arg) as pipeline:
                if pipeline.exists(self.key):  # Available since Redis 1.0.0
                    raise KeyExistsError(self.redis, self.key)
                self._populate(pipeline, arg, **kwargs)

    def _populate(self,
                  pipeline: Pipeline,
                  arg: InitArg = tuple(),
                  **kwargs: JSONTypes,
                  ) -> None:
        if isinstance(arg, collections.abc.Mapping):
            arg = arg.items()
        items = itertools.chain(arg, kwargs.items())
        dict_ = dict(items)
        encoded_dict = self.__encode_dict(dict_)
        if encoded_dict:
            if len(encoded_dict) > 1:
                warnings.warn(
                    cast(str, InefficientAccessWarning.__doc__),
                    InefficientAccessWarning,
                )
            pipeline.multi()  # Available since Redis 1.2.0
            # Available since Redis 2.0.0:
            pipeline.hset(self.key, mapping=encoded_dict)  # type: ignore

    # Preserve the Open-Closed Principle with name mangling.
    #   https://youtu.be/miGolgp9xq8?t=2086
    #   https://stackoverflow.com/a/38534939
    __populate = _populate

    def _encode_dict(self, dict_: Mapping[JSONTypes, JSONTypes]) -> Dict[str, str]:
        encoded_dict = {
            self._encode(key): self._encode(value)
            for key, value in dict_.items()
        }
        return encoded_dict

    __encode_dict = _encode_dict

    # Methods required by collections.abc.MutableMapping:

    def __getitem__(self, key: JSONTypes) -> JSONTypes:
        'd.__getitem__(key) <==> d[key].  O(1)'
        encoded_key = self._encode(key)
        encoded_value = self.redis.hget(self.key, encoded_key)  # Available since Redis 2.0.0
        if encoded_value is None:
            raise KeyError(key)
        value = self._decode(encoded_value)
        return value

    def __setitem__(self, key: JSONTypes, value: JSONTypes) -> None:
        'd.__setitem__(key, value) <==> d[key] = value.  O(1)'
        encoded_key = self._encode(key)
        encoded_value = self._encode(value)
        self.redis.hset(self.key, encoded_key, encoded_value)  # Available since Redis 2.0.0

    def __delitem__(self, key: JSONTypes) -> None:
        'd.__delitem__(key) <==> del d[key].  O(1)'
        encoded_key = self._encode(key)
        if not self.redis.hdel(self.key, encoded_key):  # Available since Redis 2.0.0
            raise KeyError(key)

    def __iter__(self) -> Generator[JSONTypes, None, None]:
        warnings.warn(
            cast(str, InefficientAccessWarning.__doc__),
            InefficientAccessWarning,
        )
        encoded_items = self.redis.hscan_iter(self.key)  # Available since Redis 2.8.0
        keys = (self._decode(key) for key, _ in encoded_items)
        yield from keys

    def __len__(self) -> int:
        'Return the number of items in the RedisDict.  O(1)'
        return self.redis.hlen(self.key)  # Available since Redis 2.0.0

    # Methods required for Raj's sanity:

    def __repr__(self) -> str:
        'Return the string representation of the RedisDict.  O(n)'
        return f'{self.__class__.__qualname__}{self.__to_dict()}'

    # Method overrides:

    # From collections.abc.MutableMapping:
    def update(self, arg: InitArg = tuple(), **kwargs: JSONTypes) -> None:  # type: ignore
        with self._watch(arg) as pipeline:
            self.__populate(pipeline, arg, **kwargs)

    # From collections.abc.Mapping:
    def __contains__(self, key: Any) -> bool:
        'd.__contains__(key) <==> key in d.  O(1)'
        try:
            encoded_key = self._encode(key)
        except TypeError:
            return False
        return self.redis.hexists(self.key, encoded_key)  # Available since Redis 2.0.0

    def to_dict(self) -> Dict[JSONTypes, JSONTypes]:
        'Convert a RedisDict into a plain Python dict.'
        encoded_items = self.redis.hgetall(self.key).items()  # Available since Redis 2.0.0
        if encoded_items:
            warnings.warn(
                cast(str, InefficientAccessWarning.__doc__),
                InefficientAccessWarning,
            )
        dict_ = {
            self._decode(encoded_key): self._decode(encoded_value)
            for encoded_key, encoded_value in encoded_items
        }
        return dict_

    __to_dict = to_dict



# --------------------------------------------------------------------------- #
#   base.py                                                                   #
#                                                                             #
#   Copyright © 2015-2022, Rajiv Bakulesh Shah, original author.              #
#                                                                             #
#   Licensed under the Apache License, Version 2.0 (the "License");           #
#   you may not use this file except in compliance with the License.          #
#   You may obtain a copy of the License at:                                  #
#       http://www.apache.org/licenses/LICENSE-2.0                            #
#                                                                             #
#   Unless required by applicable law or agreed to in writing, software       #
#   distributed under the License is distributed on an "AS IS" BASIS,         #
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  #
#   See the License for the specific language governing permissions and       #
#   limitations under the License.                                            #
# --------------------------------------------------------------------------- #


# TODO: When we drop support for Python 3.9, remove the following import.  We
# only need it for X | Y union type annotations as of 2022-01-29.
from __future__ import annotations

import abc
import collections
import contextlib
import json
import os
import uuid
import warnings
from typing import Any
from typing import AnyStr
from typing import ClassVar
from typing import ContextManager
from typing import FrozenSet
from typing import Generator
from typing import Iterable
from typing import List
from typing import Tuple
from typing import cast

from redis import Redis
from redis import RedisError
from redis.asyncio import Redis as AIORedis  # type: ignore
from redis.client import Pipeline
# TODO: When we drop support for Python 3.7, change the following imports to:
#   from typing import Final
#   from typing import Protocol
#   from typing import final
from typing_extensions import Final
from typing_extensions import Protocol
from typing_extensions import final

from .annotations import JSONTypes
from .exceptions import InefficientAccessWarning
from .exceptions import QuorumIsImpossible
from .exceptions import RandomKeyError
from .monkey import logger


_default_url: Final[str] = os.environ.get('REDIS_URL', 'redis://localhost:6379/0')
_default_redis: Final[Redis] = Redis.from_url(_default_url, socket_timeout=1)


def random_key(*,
               redis: Redis,
               prefix: str = 'pottery:',
               num_tries: int = 3,
               ) -> str:
    'Find/return a random key that does not exist in the Redis instance.'
    if not isinstance(num_tries, int):
        raise TypeError('num_tries must be an int >= 0')
    if num_tries < 0:
        raise ValueError('num_tries must be an int >= 0')
    if num_tries == 0:
        raise RandomKeyError(redis)

    uuid4 = str(uuid.uuid4())
    key = prefix + uuid4
    if redis.exists(key):  # Available since Redis 1.0.0
        key = random_key(
            redis=redis,
            prefix=prefix,
            num_tries=num_tries-1,
        )
    return key


def _connection_args(redis: Redis) -> Tuple[str, int, int]:
    return (
        redis.connection_pool.connection_kwargs['host'],
        redis.connection_pool.connection_kwargs.get('port', 6379),
        redis.connection_pool.connection_kwargs.get('db', 0),
    )


class _Common:
    'Mixin class that implements self.redis and self.key properties.'

    _RANDOM_KEY_PREFIX: ClassVar[str] = 'pottery:'

    def __init__(self, *, redis: Redis | None = None, key: str = '') -> None:
        self.redis = redis or _default_redis
        self.key = key or self.__random_key()

    def __del__(self) -> None:
        if self.key.startswith(self._RANDOM_KEY_PREFIX):
            self.redis.unlink(self.key)  # Available since Redis 4.0.0
            logger.warning(
                "Deleted tmp <%s key='%s'> (instance is about to be destroyed)",
                self.__class__.__qualname__,
                self.key,
            )

    @property
    def redis(self) -> Redis:
        # Preserve the Open-Closed Principle with name mangling.
        #   https://youtu.be/miGolgp9xq8?t=2086
        #   https://stackoverflow.com/a/38534939
        return self.__redis

    @redis.setter
    def redis(self, value: Redis | None) -> None:
        self.__redis = value or _default_redis

    @property
    def key(self) -> str:
        return self.__key  # type: ignore

    @key.setter
    def key(self, value: str) -> None:
        self.__key = value or self.__random_key()

    def _random_key(self) -> str:
        key = random_key(redis=self.redis, prefix=self._RANDOM_KEY_PREFIX)
        logger.warning(
            "Self-assigning tmp key <%s key='%s'>",
            self.__class__.__qualname__,
            key,
        )
        return key

    __random_key = _random_key


class _Encodable:
    'Mixin class that implements JSON encoding and decoding.'

    @final
    @staticmethod
    def _encode(decoded_value: JSONTypes) -> str:
        encoded_value = json.dumps(decoded_value, sort_keys=True)
        return encoded_value

    @final
    @staticmethod
    def _decode(encoded_value: AnyStr) -> JSONTypes:
        if isinstance(encoded_value, bytes):
            string = encoded_value.decode()
        else:
            string = encoded_value
        decoded_value: JSONTypes = json.loads(string)
        return decoded_value


class _HasRedisClientAndKey(Protocol):
    @property
    def redis(self) -> Redis:  # pragma: no cover
        ...

    @property
    def key(self) -> str:  # pragma: no cover
        ...


class _Clearable:
    'Mixin class that implements clearing (emptying) a Redis-backed collection.'
    def clear(self: _HasRedisClientAndKey) -> None:
        'Remove the elements in a Redis-backed container.  O(n)'
        self.redis.unlink(self.key)  # Available since Redis 4.0.0


class _Pipelined(abc.ABC):
    @property
    @abc.abstractmethod
    def redis(self) -> Redis:
        'Redis client.'

    @property
    @abc.abstractmethod
    def key(self) -> str:
        'Redis key.'

    @final
    @contextlib.contextmanager
    def __watch_keys(self, *keys: str) -> Generator[Pipeline, None, None]:
        with self.redis.pipeline() as pipeline:
            pipeline.watch(*keys)  # Available since Redis 2.2.0
            try:
                yield pipeline
            except Exception as error:
                logger.warning(
                    'Caught %s; aborting pipeline of %d commands',
                    error.__class__.__qualname__,
                    len(pipeline),
                )
                raise
            else:
                logger.info(
                    'Running EXEC on pipeline of %d commands',
                    len(pipeline),
                )
                pipeline.execute()  # Available since Redis 1.2.0

    @final
    def __context_managers(self,
                           *others: Any,
                           ) -> Generator[ContextManager[Pipeline], None, None]:
        redis_clients = collections.defaultdict(list)
        for container in (self, *others):
            if isinstance(container, _Pipelined):
                connection = _connection_args(container.redis)
                redis_clients[connection].append(container)
        for containers in redis_clients.values():
            keys = (container.key for container in containers)
            pipeline = containers[0].__watch_keys(*keys)
            yield pipeline

    @final
    @contextlib.contextmanager
    def _watch(self, *others: Any) -> Generator[Pipeline, None, None]:
        'Watch self and others, and yield a Redis pipeline.'
        pipelines = []
        with contextlib.ExitStack() as stack:
            for context_manager in self.__context_managers(*others):
                pipeline = stack.enter_context(context_manager)
                pipelines.append(pipeline)
            yield pipelines[0]


class _Comparable(abc.ABC):
    'Mixin class that implements equality testing for Redis-backed collections.'

    @property
    @abc.abstractmethod
    def redis(self) -> Redis:
        'Redis client.'

    @property
    @abc.abstractmethod
    def key(self) -> str:
        'Redis key.'

    @abc.abstractmethod
    @contextlib.contextmanager
    def _watch(self, *others: Any) -> Generator[Pipeline, None, None]:
        'Watch self and others, and yield a Redis pipeline.'

    @final
    def _same_redis(self, *others: Any) -> bool:
        for other in others:
            if not isinstance(other, _Comparable):
                return False
            if _connection_args(self.redis) != _connection_args(other.redis):
                return False
        return True

    __same_redis = _same_redis

    def __eq__(self, other: Any) -> bool:
        if self is other:
            return True
        if self.__same_redis(other) and self.key == other.key:
            return True

        warnings.warn(
            cast(str, InefficientAccessWarning.__doc__),
            InefficientAccessWarning,
        )
        with self._watch(other):
            equals = super().__eq__(other)
        if equals is NotImplemented:
            equals = False
        return equals


class Container(_Common, _Encodable, _Clearable, _Pipelined, _Comparable):
    'Base class for Redis-backed collections.'


class Iterable_(abc.ABC):
    'Mixin class that implements iterating over a Redis-backed collection.'

    @abc.abstractmethod
    def __iter__(self) -> Generator[JSONTypes, None, None]:
        raise NotImplementedError


class Primitive(abc.ABC):
    'Base class for Redis-backed distributed primitives.'

    @property
    @abc.abstractmethod
    def _KEY_PREFIX(self) -> str:
        'Redis key prefix/namespace.'

    __slots__ = ('__key', 'masters', 'raise_on_redis_errors')

    _DEFAULT_MASTERS: ClassVar[FrozenSet[Redis]] = frozenset({_default_redis})

    def __init__(self,
                 *,
                 key: str,
                 masters: Iterable[Redis] = frozenset(),
                 raise_on_redis_errors: bool = False,
                 ) -> None:
        self.key = key
        self.masters = frozenset(masters) or self._DEFAULT_MASTERS
        self.raise_on_redis_errors = raise_on_redis_errors

    @property
    def key(self) -> str:
        return self.__key

    @key.setter
    def key(self, value: str) -> None:
        self.__key = f'{self._KEY_PREFIX}:{value}'

    def _check_enough_masters_up(self,
                                 raise_on_redis_errors: bool | None,
                                 redis_errors: List[RedisError],
                                 ) -> None:
        if raise_on_redis_errors is None:
            raise_on_redis_errors = self.raise_on_redis_errors
        if raise_on_redis_errors and len(redis_errors) > len(self.masters) // 2:
            raise QuorumIsImpossible(
                self.key,
                self.masters,
                redis_errors=redis_errors,
            )


class AIOPrimitive(Primitive):
    __slots__: Tuple[str, ...] = tuple()

    _DEFAULT_MASTERS: ClassVar[FrozenSet[AIORedis]] = frozenset()  # type: ignore



# --------------------------------------------------------------------------- #
#   aionextid.py                                                              #
#                                                                             #
#   Copyright © 2015-2022, Rajiv Bakulesh Shah, original author.              #
#                                                                             #
#   Licensed under the Apache License, Version 2.0 (the "License");           #
#   you may not use this file except in compliance with the License.          #
#   You may obtain a copy of the License at:                                  #
#       http://www.apache.org/licenses/LICENSE-2.0                            #
#                                                                             #
#   Unless required by applicable law or agreed to in writing, software       #
#   distributed under the License is distributed on an "AS IS" BASIS,         #
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  #
#   See the License for the specific language governing permissions and       #
#   limitations under the License.                                            #
# --------------------------------------------------------------------------- #
'Asynchronous distributed Redis-powered monotonically increasing ID generator.'


# TODO: Remove the following import after deferred evaluation of annotations
# because the default.
#   1. https://docs.python.org/3/whatsnew/3.7.html#whatsnew37-pep563
#   2. https://www.python.org/dev/peps/pep-0563/
#   3. https://www.python.org/dev/peps/pep-0649/
from __future__ import annotations

import asyncio
import contextlib
from typing import ClassVar
from typing import Iterable

from redis import RedisError
from redis.asyncio import Redis as AIORedis  # type: ignore

from .base import AIOPrimitive
from .base import logger
from .exceptions import QuorumNotAchieved
from .nextid import NextID
from .nextid import Scripts


class AIONextID(Scripts, AIOPrimitive):
    'Async distributed Redis-powered monotonically increasing ID generator.'

    __slots__ = ('num_tries',)

    _KEY_PREFIX: ClassVar[str] = NextID._KEY_PREFIX

    def __init__(self,  # type: ignore
                 *,
                 key: str = 'current',
                 masters: Iterable[AIORedis] = frozenset(),
                 num_tries: int = NextID._NUM_TRIES,
                 ) -> None:
        'Initialize an AIONextID ID generator.'
        super().__init__(key=key, masters=masters)
        self.num_tries = num_tries

    def __aiter__(self) -> AIONextID:
        return self  # pragma: no cover

    async def __anext__(self) -> int:
        for _ in range(self.num_tries):
            with contextlib.suppress(QuorumNotAchieved):
                next_id = await self.__get_current_ids() + 1
                await self.__set_current_ids(next_id)
                return next_id
        raise QuorumNotAchieved(self.key, self.masters)

    async def __get_current_id(self, master: AIORedis) -> int:  # type: ignore
        current_id: int = await master.get(self.key)
        return current_id

    async def __set_current_id(self, master: AIORedis, value: int) -> bool:  # type: ignore
        current_id: int | None = await self._set_id_script(  # type: ignore
            keys=(self.key,),
            args=(value,),
            client=master,
        )
        return current_id == value

    async def __reset_current_id(self, master: AIORedis) -> int:  # type: ignore
        await master.delete(self.key)

    async def __get_current_ids(self) -> int:
        current_ids, redis_errors = [], []
        coros = [self.__get_current_id(master) for master in self.masters]
        for coro in asyncio.as_completed(coros):  # type: ignore
            try:
                current_id = int(await coro or b'0')
            except RedisError as error:
                redis_errors.append(error)
                logger.exception(
                    '%s.__get_current_ids() caught %s',
                    self.__class__.__qualname__,
                    error.__class__.__qualname__,
                )
            else:
                current_ids.append(current_id)
        if len(current_ids) > len(self.masters) // 2:
            return max(current_ids)
        raise QuorumNotAchieved(
            self.key,
            self.masters,
            redis_errors=redis_errors,
        )

    async def __set_current_ids(self, value: int) -> None:
        num_masters_set, redis_errors = 0, []
        coros = [self.__set_current_id(master, value) for master in self.masters]
        for coro in asyncio.as_completed(coros):
            try:
                num_masters_set += await coro
            except RedisError as error:
                redis_errors.append(error)
                logger.exception(
                    '%s.__set_current_ids() caught %s',
                    self.__class__.__qualname__,
                    error.__class__.__qualname__,
                )
        if num_masters_set > len(self.masters) // 2:
            return
        raise QuorumNotAchieved(
            self.key,
            self.masters,
            redis_errors=redis_errors,
        )

    async def reset(self) -> None:
        num_masters_reset, redis_errors = 0, []
        coros = [self.__reset_current_id(master) for master in self.masters]
        for coro in asyncio.as_completed(coros):
            try:
                await coro
            except RedisError as error:
                redis_errors.append(error)
                logger.exception(
                    '%s.reset() caught %s',
                    self.__class__.__qualname__,
                    error.__class__.__qualname__,
                )
            else:
                num_masters_reset += 1
        if num_masters_reset > len(self.masters) // 2:
            return
        raise QuorumNotAchieved(
            self.key,
            self.masters,
            redis_errors=redis_errors,
        )

    def __repr__(self) -> str:
        return f'<{self.__class__.__qualname__} key={self.key}>'



# --------------------------------------------------------------------------- #
#   list.py                                                                   #
#                                                                             #
#   Copyright © 2015-2022, Rajiv Bakulesh Shah, original author.              #
#                                                                             #
#   Licensed under the Apache License, Version 2.0 (the "License");           #
#   you may not use this file except in compliance with the License.          #
#   You may obtain a copy of the License at:                                  #
#       http://www.apache.org/licenses/LICENSE-2.0                            #
#                                                                             #
#   Unless required by applicable law or agreed to in writing, software       #
#   distributed under the License is distributed on an "AS IS" BASIS,         #
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  #
#   See the License for the specific language governing permissions and       #
#   limitations under the License.                                            #
# --------------------------------------------------------------------------- #


# TODO: Remove the following import after deferred evaluation of annotations
# because the default.
#   1. https://docs.python.org/3/whatsnew/3.7.html#whatsnew37-pep563
#   2. https://www.python.org/dev/peps/pep-0563/
#   3. https://www.python.org/dev/peps/pep-0649/
from __future__ import annotations

import collections.abc
import functools
import itertools
import uuid
import warnings
from typing import Any
from typing import Callable
from typing import Iterable
from typing import List
from typing import cast

from redis import Redis
from redis import ResponseError
from redis.client import Pipeline
# TODO: When we drop support for Python 3.7, change the following import to:
#   from typing import final
from typing_extensions import final

from .annotations import F
from .annotations import JSONTypes
from .base import Container
from .exceptions import InefficientAccessWarning
from .exceptions import KeyExistsError


def _raise_on_error(func: F) -> Callable[[F], F]:
    @functools.wraps(func)
    def wrapper(*args: Any, **kwargs: Any) -> Any:
        try:
            return func(*args, **kwargs)
        except ResponseError as error:
            raise IndexError('list assignment index out of range') from error
    return wrapper


class RedisList(Container, collections.abc.MutableSequence):
    'Redis-backed container compatible with Python lists.'

    _ALLOWED_TO_EQUAL: type = list

    def __slice_to_indices(self, slice_or_index: slice | int) -> range:
        if isinstance(slice_or_index, slice):
            start, stop, step = slice_or_index.indices(len(self))
        elif isinstance(slice_or_index, int):
            start = slice_or_index
            stop = slice_or_index + 1
            step = 1
        else:
            raise TypeError(
                'list indices must be integers or slices, '
                f'not {slice_or_index.__class__.__qualname__}'
            )
        indices = range(start, stop, step)
        return indices

    def __init__(self,
                 iterable: Iterable[JSONTypes] = tuple(),
                 *,
                 redis: Redis | None = None,
                 key: str = '',
                 ) -> None:
        'Initialize the RedisList.  O(n)'
        super().__init__(redis=redis, key=key)
        if iterable:
            with self._watch(iterable) as pipeline:
                if pipeline.exists(self.key):  # Available since Redis 1.0.0
                    raise KeyExistsError(pipeline, self.key)
                self._populate(pipeline, iterable)

    def _populate(self,
                  pipeline: Pipeline,
                  iterable: Iterable[JSONTypes] = tuple(),
                  ) -> None:
        encoded_values = [self._encode(value) for value in iterable]
        if encoded_values:
            if len(encoded_values) > 1:
                warnings.warn(
                    cast(str, InefficientAccessWarning.__doc__),
                    InefficientAccessWarning,
                )
            pipeline.multi()  # Available since Redis 1.2.0
            pipeline.rpush(self.key, *encoded_values)  # Available since Redis 1.0.0

    # Methods required by collections.abc.MutableSequence:

    def __getitem__(self, index: slice | int) -> Any:
        'l.__getitem__(index) <==> l[index].  O(n)'
        with self._watch() as pipeline:
            if isinstance(index, slice):
                # Python's list API requires us to get elements by slice (a
                # start index, a stop index, and a step).  Redis supports only
                # getting elements by start and stop (no step).  So our
                # workaround is to get all of the elements between start and
                # stop from Redis, then discard the ones between step in Python
                # code.  More info:
                #   http://redis.io/commands/lrange
                warnings.warn(
                    cast(str, InefficientAccessWarning.__doc__),
                    InefficientAccessWarning,
                )
                indices = self.__slice_to_indices(index)
                if indices.step >= 0:
                    start, stop = indices.start, indices.stop-1
                else:
                    start, stop = indices.stop+1, indices.start
                pipeline.multi()  # Available since Redis 1.2.0
                pipeline.lrange(self.key, start, stop)  # Available since Redis 1.0.0
                encoded_values = pipeline.execute()[0]  # Available since Redis 1.2.0
                encoded_values = encoded_values[::index.step]
                value: List[JSONTypes] | JSONTypes = [
                    self._decode(value) for value in encoded_values
                ]
            else:
                index = self.__slice_to_indices(index).start
                len_ = cast(int, pipeline.llen(self.key))  # Available since Redis 1.0.0
                if index not in {-1, 0, len_-1}:
                    warnings.warn(
                        cast(str, InefficientAccessWarning.__doc__),
                        InefficientAccessWarning,
                    )
                encoded_value = pipeline.lindex(self.key, index)  # Available since Redis 1.0.0
                if encoded_value is None:
                    raise IndexError('list index out of range')
                value = self._decode(cast(bytes, encoded_value))
        return value

    @_raise_on_error
    def __setitem__(self, index: slice | int, value: JSONTypes) -> None:  # type: ignore
        'l.__setitem__(index, value) <==> l[index] = value.  O(n)'
        with self._watch() as pipeline:
            if isinstance(index, slice):
                warnings.warn(
                    cast(str, InefficientAccessWarning.__doc__),
                    InefficientAccessWarning,
                )
                if not isinstance(value, Iterable):
                    raise TypeError('can only assign an iterable')
                encoded_values = [self._encode(value) for value in value]
                indices = self.__slice_to_indices(index)
                pipeline.multi()  # Available since Redis 1.2.0
                for index, encoded_value in zip(indices, encoded_values):
                    pipeline.lset(self.key, index, encoded_value)  # Available since Redis 1.0.0
                indices, num = indices[len(encoded_values):], 0
                for index in indices:
                    pipeline.lset(self.key, index, 0)  # Available since Redis 1.0.0
                    num += 1
                if num:
                    pipeline.lrem(self.key, num, 0)  # Available since Redis 1.0.0
            else:
                index = self.__slice_to_indices(index).start
                len_ = cast(int, pipeline.llen(self.key))  # Available since Redis 1.0.0
                if index not in {-1, 0, len_-1}:
                    warnings.warn(
                        cast(str, InefficientAccessWarning.__doc__),
                        InefficientAccessWarning,
                    )
                pipeline.multi()  # Available since Redis 1.2.0
                encoded_value = self._encode(value)
                pipeline.lset(self.key, index, encoded_value)  # Available since Redis 1.0.0

    @_raise_on_error
    def __delitem__(self, index: slice | int) -> None:  # type: ignore
        'l.__delitem__(index) <==> del l[index].  O(n)'
        with self._watch() as pipeline:
            self.__delete(pipeline, index)

    # Preserve the Open-Closed Principle with name mangling.
    #   https://youtu.be/miGolgp9xq8?t=2086
    #   https://stackoverflow.com/a/38534939
    def __delete(self, pipeline: Pipeline, index: slice | int) -> None:
        # Python's list API requires us to delete an element by *index.*  Redis
        # supports only deleting an element by *value.*  So our workaround is
        # to set l[index] to a UUID4, then to delete the value UUID4.  More
        # info:
        #   http://redis.io/commands/lrem
        warnings.warn(
            cast(str, InefficientAccessWarning.__doc__),
            InefficientAccessWarning,
        )
        indices, num = self.__slice_to_indices(index), 0
        uuid4 = str(uuid.uuid4())
        pipeline.multi()  # Available since Redis 1.2.0
        for index in indices:
            pipeline.lset(self.key, index, uuid4)  # Available since Redis 1.0.0
            num += 1
        if num:
            pipeline.lrem(self.key, num, uuid4)  # Available since Redis 1.0.0

    def __len__(self) -> int:
        'Return the number of items in the RedisList.  O(1)'
        return self.redis.llen(self.key)  # Available since Redis 1.0.0

    def insert(self, index: int, value: JSONTypes) -> None:
        'Insert an element into the RedisDeque before the given index.  O(n)'
        with self._watch() as pipeline:
            self.__insert(index, value, pipeline=pipeline)

    def _insert(self,
                index: int,
                value: JSONTypes,
                *,
                pipeline: Pipeline,
                ) -> None:
        encoded_value = self._encode(value)
        current_length = cast(int, pipeline.llen(self.key))  # Available since Redis 1.0.0
        if 0 < index < current_length:
            # Python's list API requires us to insert an element before the
            # given *index.*  Redis supports only inserting an element before a
            # given (pivot) *value.*  So our workaround is to set the pivot
            # value to a UUID4, then to insert the desired value before the
            # UUID4, then to set the value UUID4 back to the original pivot
            # pivot value.  More info:
            #   http://redis.io/commands/linsert
            warnings.warn(
                cast(str, InefficientAccessWarning.__doc__),
                InefficientAccessWarning,
            )
            uuid4 = str(uuid.uuid4())
            pivot = cast(bytes, pipeline.lindex(self.key, index))
            pipeline.multi()  # Available since Redis 1.2.0
            pipeline.lset(self.key, index, uuid4)  # Available since Redis 1.0.0
            pipeline.linsert(self.key, 'BEFORE', uuid4, encoded_value)
            pipeline.lset(self.key, index+1, pivot)  # Available since Redis 1.0.0
        else:
            pipeline.multi()  # Available since Redis 1.2.0
            push_method = pipeline.lpush if index <= 0 else pipeline.rpush  # Available since Redis 1.0.0
            push_method(self.key, encoded_value)

    __insert = _insert

    # Methods required for Raj's sanity:

    def sort(self, *, key: str | None = None, reverse: bool = False) -> None:
        'Sort the RedisList in place.  O(n)'
        if key is not None:
            raise NotImplementedError('sorting by key not implemented')
        warnings.warn(
            cast(str, InefficientAccessWarning.__doc__),
            InefficientAccessWarning,
        )
        self.redis.sort(self.key, desc=reverse, store=self.key)  # Available since Redis 1.0.0

    def __eq__(self, other: Any) -> bool:
        if type(other) not in {self.__class__, self._ALLOWED_TO_EQUAL}:
            return False
        if self is other:
            return True
        if self._same_redis(other) and self.key == other.key:
            return True

        with self._watch(other):
            if len(self) != len(other):
                return False

            # other is a list or RedisList too, and self and other are the same
            # length.  Make Python lists out of self and other, and compare
            # those lists.
            warnings.warn(
                cast(str, InefficientAccessWarning.__doc__),
                InefficientAccessWarning,
            )
            self_as_list = self.__to_list()
            if isinstance(other, self.__class__):
                other_as_list = other.to_list()
            else:
                other_as_list = list(other)
            return self_as_list == other_as_list

    def __add__(self, other: List[JSONTypes]) -> RedisList:
        'Append the items in other to the RedisList.  O(n)'
        warnings.warn(
            cast(str, InefficientAccessWarning.__doc__),
            InefficientAccessWarning,
        )
        with self._watch(other):
            iterable = itertools.chain(self, other)
            return self.__class__(iterable, redis=self.redis)

    def __repr__(self) -> str:
        'Return the string representation of the RedisList.  O(n)'
        warnings.warn(
            cast(str, InefficientAccessWarning.__doc__),
            InefficientAccessWarning,
        )
        return self.__class__.__qualname__ + str(self.__to_list())

    # Method overrides:

    # From collections.abc.MutableSequence:
    def append(self, value: JSONTypes) -> None:
        'Add an element to the right side of the RedisList.  O(1)'
        self.__extend((value,))

    # From collections.abc.MutableSequence:
    def extend(self, values: Iterable[JSONTypes]) -> None:
        'Extend the RedisList by appending elements from the iterable.  O(1)'
        encoded_values = (self._encode(value) for value in values)
        self.redis.rpush(self.key, *encoded_values)  # Available since Redis 1.0.0

    __extend = extend

    # From collections.abc.MutableSequence:
    def pop(self, index: int | None = None) -> JSONTypes:
        with self._watch() as pipeline:
            len_ = len(self)
            if index and index >= len_:
                raise IndexError('pop index out of range')
            elif index in {0, None, len_-1, -1}:
                pop_method = 'lpop' if index == 0 else 'rpop'
                pipeline.multi()  # Available since Redis 1.2.0
                getattr(pipeline, pop_method)(self.key)  # Available since Redis 1.0.0
                encoded_value = pipeline.execute()[0]  # Available since Redis 1.2.0
                if encoded_value is None:
                    raise IndexError(
                        f'pop from an empty {self.__class__.__qualname__}'
                    )
                return self._decode(encoded_value)
            else:
                value: JSONTypes = self[cast(int, index)]
                self.__delete(pipeline, cast(int, index))
                return value

    # From collections.abc.MutableSequence:
    def remove(self, value: JSONTypes) -> None:
        'Remove first occurrence of value.  O(n)'
        warnings.warn(
            cast(str, InefficientAccessWarning.__doc__),
            InefficientAccessWarning,
        )
        with self._watch() as pipeline:
            for index, element in enumerate(self):
                if element == value:
                    self.__delete(pipeline, index)
                    return
        class_name = self.__class__.__qualname__
        raise ValueError(f'{class_name}.remove(x): x not in {class_name}')

    @final
    def to_list(self) -> List[JSONTypes]:
        'Convert the RedisList to a Python list.  O(n)'
        encoded = self.redis.lrange(self.key, 0, -1)  # Available since Redis 1.0.0
        if encoded:
            warnings.warn(
                cast(str, InefficientAccessWarning.__doc__),
                InefficientAccessWarning,
            )
        values = [self._decode(value) for value in encoded]
        return values

    __to_list = to_list



# --------------------------------------------------------------------------- #
#   deque.py                                                                  #
#                                                                             #
#   Copyright © 2015-2022, Rajiv Bakulesh Shah, original author.              #
#                                                                             #
#   Licensed under the Apache License, Version 2.0 (the "License");           #
#   you may not use this file except in compliance with the License.          #
#   You may obtain a copy of the License at:                                  #
#       http://www.apache.org/licenses/LICENSE-2.0                            #
#                                                                             #
#   Unless required by applicable law or agreed to in writing, software       #
#   distributed under the License is distributed on an "AS IS" BASIS,         #
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  #
#   See the License for the specific language governing permissions and       #
#   limitations under the License.                                            #
# --------------------------------------------------------------------------- #


# TODO: When we drop support for Python 3.9, remove the following import.  We
# only need it for X | Y union type annotations as of 2022-01-29.
from __future__ import annotations

import collections
import warnings
from typing import Iterable
from typing import Tuple
from typing import cast

from redis import Redis
from redis.client import Pipeline

from .annotations import JSONTypes
from .exceptions import InefficientAccessWarning
from .list import RedisList


class RedisDeque(RedisList, collections.deque):  # type: ignore
    'Redis-backed container compatible with collections.deque.'

    # Overrides:

    _ALLOWED_TO_EQUAL = collections.deque

    def __init__(self,
                 iterable: Iterable[JSONTypes] = tuple(),
                 maxlen: int | None = None,
                 *,
                 redis: Redis | None = None,
                 key: str = '',
                 ) -> None:
        'Initialize the RedisDeque.  O(n)'
        if maxlen is not None and not isinstance(maxlen, int):
            raise TypeError('an integer is required')
        self._maxlen = maxlen
        super().__init__(iterable, redis=redis, key=key)
        if not iterable and self.maxlen is not None and len(self) > self.maxlen:
            raise IndexError(
                f'persistent {self.__class__.__qualname__} beyond its maximum size'
            )

    def _populate(self,
                  pipeline: Pipeline,
                  iterable: Iterable[JSONTypes] = tuple(),
                  ) -> None:
        if self.maxlen is not None:
            if self.maxlen:
                iterable = tuple(iterable)[-self.maxlen:]
            else:
                # self.maxlen == 0.  Populate the RedisDeque with an empty
                # iterable.
                iterable = tuple()
        super()._populate(pipeline, iterable)

    @property
    def maxlen(self) -> int | None:
        return self._maxlen

    @maxlen.setter
    def maxlen(self, value: int) -> None:
        raise AttributeError(
            f"attribute 'maxlen' of '{self.__class__.__qualname__}' objects is not "
            'writable'
        )

    def insert(self, index: int, value: JSONTypes) -> None:
        'Insert an element into the RedisDeque before the given index.  O(n)'
        with self._watch() as pipeline:
            current_length = cast(int, pipeline.llen(self.key))  # Available since Redis 1.0.0
            if self.maxlen is not None and current_length >= self.maxlen:
                raise IndexError(
                    f'{self.__class__.__qualname__} already at its maximum size'
                )
            super()._insert(index, value, pipeline=pipeline)

    def append(self, value: JSONTypes) -> None:
        'Add an element to the right side of the RedisDeque.  O(1)'
        self.__extend((value,), right=True)

    def appendleft(self, value: JSONTypes) -> None:
        'Add an element to the left side of the RedisDeque.  O(1)'
        self.__extend((value,), right=False)

    def extend(self, values: Iterable[JSONTypes]) -> None:
        'Extend the RedisDeque by appending elements from the iterable.  O(1)'
        self.__extend(values, right=True)

    def extendleft(self, values: Iterable[JSONTypes]) -> None:
        '''Extend the RedisDeque by prepending elements from the iterable.  O(1)

        Note the order in which the elements are prepended from the iterable:

            >>> d = RedisDeque()
            >>> d.extendleft('abc')
            >>> d
            RedisDeque(['c', 'b', 'a'])
        '''
        self.__extend(values, right=False)

    def __extend(self,
                 values: Iterable[JSONTypes],
                 *,
                 right: bool = True,
                 ) -> None:
        with self._watch(values) as pipeline:
            push_method_name = 'rpush' if right else 'lpush'
            encoded_values = [self._encode(value) for value in values]
            len_ = cast(int, pipeline.llen(self.key)) + len(encoded_values)  # Available since Redis 1.0.0
            trim_indices: Tuple[int, int] | Tuple = tuple()
            if self.maxlen is not None and len_ >= self.maxlen:
                trim_indices = (len_-self.maxlen, len_) if right else (0, self.maxlen-1)

            pipeline.multi()  # Available since Redis 1.2.0
            push_method = getattr(pipeline, push_method_name)
            push_method(self.key, *encoded_values)
            if trim_indices:
                pipeline.ltrim(self.key, *trim_indices)  # Available since Redis 1.0.0

    def pop(self) -> JSONTypes:  # type: ignore
        return super().pop()

    def popleft(self) -> JSONTypes:
        return super().pop(0)

    def rotate(self, n: int = 1) -> None:
        '''Rotate the RedisDeque n steps to the right (default n=1).  O(n)

        If n is negative, rotates left.
        '''
        if not isinstance(n, int):
            raise TypeError(
                f"'{n.__class__.__qualname__}' object cannot be interpreted "
                'as an integer'
            )
        if n == 0:
            # Rotating 0 steps is a no-op.
            return

        with self._watch() as pipeline:
            if not self:
                # Rotating an empty RedisDeque is a no-op.
                return

            push_method_name = 'lpush' if n > 0 else 'rpush'  # Available since Redis 1.0.0
            values = self[-n:][::-1] if n > 0 else self[:-n]
            encoded_values = (self._encode(value) for value in values)
            trim_indices = (0, len(self)-1) if n > 0 else (-n, len(self)-1-n)

            pipeline.multi()  # Available since Redis 1.2.0
            push_method = getattr(pipeline, push_method_name)
            push_method(self.key, *encoded_values)
            pipeline.ltrim(self.key, *trim_indices)  # Available since Redis 1.0.0

    # Methods required for Raj's sanity:

    def __bool__(self) -> bool:
        'Whether the RedisDeque contains any elements.  O(1)'
        return bool(len(self))

    def __repr__(self) -> str:
        'Return the string representation of the RedisDeque.  O(n)'
        warnings.warn(
            cast(str, InefficientAccessWarning.__doc__),
            InefficientAccessWarning,
        )
        encoded_values = self.redis.lrange(self.key, 0, -1)  # Available since Redis 1.0.0
        values = [self._decode(value) for value in encoded_values]
        repr = self.__class__.__qualname__ + '(' + str(values)
        if self.maxlen is not None:
            repr += f', maxlen={self.maxlen}'
        repr += ')'
        return repr



# --------------------------------------------------------------------------- #
#   redlock.py                                                                #
#                                                                             #
#   Copyright © 2015-2022, Rajiv Bakulesh Shah, original author.              #
#                                                                             #
#   Licensed under the Apache License, Version 2.0 (the "License");           #
#   you may not use this file except in compliance with the License.          #
#   You may obtain a copy of the License at:                                  #
#       http://www.apache.org/licenses/LICENSE-2.0                            #
#                                                                             #
#   Unless required by applicable law or agreed to in writing, software       #
#   distributed under the License is distributed on an "AS IS" BASIS,         #
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  #
#   See the License for the specific language governing permissions and       #
#   limitations under the License.                                            #
# --------------------------------------------------------------------------- #
'''Distributed Redis-powered lock.

This algorithm safely and reliably provides a mutually-exclusive locking
primitive to protect a resource shared across threads, processes, and even
machines, without a single point of failure.

Rationale and algorithm description:
    http://redis.io/topics/distlock

Reference implementations:
    https://github.com/antirez/redlock-rb
    https://github.com/SPSCommerce/redlock-py

Lua scripting:
    https://github.com/andymccurdy/redis-py#lua-scripting
'''


# TODO: Remove the following import after deferred evaluation of annotations
# because the default.
#   1. https://docs.python.org/3/whatsnew/3.7.html#whatsnew37-pep563
#   2. https://www.python.org/dev/peps/pep-0563/
#   3. https://www.python.org/dev/peps/pep-0649/
from __future__ import annotations

import concurrent.futures
import contextlib
import functools
import math
import random
import time
import uuid
from types import TracebackType
from typing import Any
from typing import Callable
from typing import ClassVar
from typing import Iterable
from typing import Tuple
from typing import Type
from typing import cast
from typing import overload

from redis import Redis
from redis import RedisError
from redis.commands.core import Script
# TODO: When we drop support for Python 3.7, change the following import to:
#   from typing import Literal
from typing_extensions import Literal

from .annotations import F
from .base import Primitive
from .base import logger
from .exceptions import ExtendUnlockedLock
from .exceptions import QuorumNotAchieved
from .exceptions import ReleaseUnlockedLock
from .exceptions import TooManyExtensions
from .executor import BailOutExecutor
from .timer import ContextTimer


class Scripts:
    '''Mixin class to define/register Lua scripts for Redis.

    Note that we only have to register these Lua scripts once -- so we do it on
    the first instantiation of Redlock.
    '''

    __slots__: Tuple[str, ...] = tuple()

    _acquired_script: ClassVar[Script | None] = None
    _extend_script: ClassVar[Script | None] = None
    _release_script: ClassVar[Script | None] = None

    def __init__(self,
                 *,
                 key: str,
                 masters: Iterable[Redis] = frozenset(),
                 raise_on_redis_errors: bool = False,
                 ) -> None:
        super().__init__(  # type: ignore
            key=key,
            masters=masters,
            raise_on_redis_errors=raise_on_redis_errors,
        )
        self.__register_acquired_script()
        self.__register_extend_script()
        self.__register_release_script()

    # Preserve the Open-Closed Principle with name mangling.
    #   https://youtu.be/miGolgp9xq8?t=2086
    #   https://stackoverflow.com/a/38534939
    def __register_acquired_script(self) -> None:
        if self._acquired_script is None:
            class_name = self.__class__.__qualname__
            logger.info('Registering %s._acquired_script', class_name)
            master = next(iter(self.masters))  # type: ignore
            # Available since Redis 2.6.0:
            self.__class__._acquired_script = master.register_script('''
                if redis.call('get', KEYS[1]) == ARGV[1] then
                    local pttl = redis.call('pttl', KEYS[1])
                    return (pttl > 0) and pttl or 0
                else
                    return 0
                end
            ''')

    def __register_extend_script(self) -> None:
        if self._extend_script is None:
            class_name = self.__class__.__qualname__
            logger.info('Registering %s._extend_script', class_name)
            master = next(iter(self.masters))  # type: ignore
            # Available since Redis 2.6.0:
            self.__class__._extend_script = master.register_script('''
                if redis.call('get', KEYS[1]) == ARGV[1] then
                    return redis.call('pexpire', KEYS[1], ARGV[2])
                else
                    return 0
                end
            ''')

    def __register_release_script(self) -> None:
        if self._release_script is None:
            class_name = self.__class__.__qualname__
            logger.info('Registering %s._release_script', class_name)
            master = next(iter(self.masters))  # type: ignore
            # Available since Redis 2.6.0:
            self.__class__._release_script = master.register_script('''
                if redis.call('get', KEYS[1]) == ARGV[1] then
                    return redis.call('del', KEYS[1])
                else
                    return 0
                end
            ''')


class Redlock(Scripts, Primitive):
    '''Distributed Redis-powered lock.

    This algorithm safely and reliably provides a mutually-exclusive locking
    primitive to protect a resource shared across threads, processes, and even
    machines, without a single point of failure.

    Rationale and algorithm description:
        http://redis.io/topics/distlock

    Usage:

        >>> from redis import Redis
        >>> redis = Redis()
        >>> printer_lock = Redlock(key='printer', masters={redis}, auto_release_time=.2)
        >>> bool(printer_lock.locked())
        False
        >>> if printer_lock.acquire():
        ...     # Critical section - print stuff here.
        ...     print('printer_lock is locked')
        printer_lock is locked
        >>> printer_lock.release()
        >>> bool(printer_lock.locked())
        False

    Redlocks time out (by default, after 10 seconds).  You should take care to
    ensure that your critical section completes well within the timeout.  The
    reasons that Redlocks time out are to preserve "liveness"
    (http://redis.io/topics/distlock#liveness-arguments) and to avoid deadlocks
    (in the event that a process dies inside a critical section before it
    releases its lock).

        >>> if printer_lock.acquire():
        ...     # Critical section - print stuff here.
        ...     time.sleep(printer_lock.auto_release_time)
        >>> bool(printer_lock.locked())
        False

    If 10 seconds isn't enough to complete executing your critical section,
    then you can specify your own timeout:

        >>> printer_lock = Redlock(key='printer', masters={redis}, auto_release_time=.2)
        >>> if printer_lock.acquire():
        ...     # Critical section - print stuff here.
        ...     time.sleep(printer_lock.auto_release_time / 2)
        >>> bool(printer_lock.locked())
        True
        >>> time.sleep(printer_lock.auto_release_time / 2)
        >>> bool(printer_lock.locked())
        False

    You can use a Redlock as a context manager:

        >>> with Redlock(key='printer', masters={redis}) as printer_lock:
        ...     # Critical section - print stuff here.
        ...     print('printer_lock is locked')
        printer_lock is locked
        >>> bool(printer_lock.locked())
        False

        >>> with printer_lock:
        ...     # Critical section - print stuff here.
        ...     print('printer_lock is locked')
        printer_lock is locked
        >>> bool(printer_lock.locked())
        False
    '''

    __slots__ = (
        'auto_release_time',
        'num_extensions',
        'context_manager_blocking',
        'context_manager_timeout',
        '_uuid',
        '_extension_num',
    )

    _KEY_PREFIX: ClassVar[str] = 'redlock'
    _AUTO_RELEASE_TIME: ClassVar[float] = 10
    _CLOCK_DRIFT_FACTOR: ClassVar[float] = 0.01
    _RETRY_DELAY: ClassVar[float] = .2
    _NUM_EXTENSIONS: ClassVar[int] = 3

    def __init__(self,
                 *,
                 key: str,
                 masters: Iterable[Redis] = frozenset(),
                 raise_on_redis_errors: bool = False,
                 auto_release_time: float = _AUTO_RELEASE_TIME,
                 num_extensions: int = _NUM_EXTENSIONS,
                 context_manager_blocking: bool = True,
                 context_manager_timeout: float = -1,
                 ) -> None:
        '''Initialize a Redlock.

        Keyword arguments:
            key -- a string that identifies your resource
            masters -- the Redis clients used to achieve quorum for this
                Redlock's state
            raise_on_redis_errors -- whether to raise the QuorumIsImplssible
                exception when too many Redis masters throw errors
            auto_release_time -- the timeout in seconds by which to
                automatically release this Redlock, unless it's already been
                released
            num_extensions -- the number of times that this Redlock's lease can
                be extended
            context_manager_blocking -- when using this Redlock as a context
                manager, whether to block when acquiring
            context_manager_timeout -- if context_manager_blocking, how long to
                wait when acquiring before giving up and raising the
                QuorumNotAchieved exception
        '''
        if not context_manager_blocking and context_manager_timeout != -1:
            raise ValueError("can't specify a timeout for a non-blocking call")

        super().__init__(
            key=key,
            masters=masters,
            raise_on_redis_errors=raise_on_redis_errors,
        )
        self.auto_release_time = auto_release_time
        self.num_extensions = num_extensions
        self.context_manager_blocking = context_manager_blocking
        self.context_manager_timeout = context_manager_timeout
        self._uuid = ''
        self._extension_num = 0

    def __acquire_master(self, master: Redis) -> bool:
        acquired = master.set(
            self.key,
            self._uuid,
            px=int(self.auto_release_time * 1000),
            nx=True,
        )
        return bool(acquired)

    def __acquired_master(self, master: Redis) -> int:
        if self._uuid:
            ttl: int = cast(Script, self._acquired_script)(
                keys=(self.key,),
                args=(self._uuid,),
                client=master,
            )
        else:
            ttl = 0
        return ttl

    def __extend_master(self, master: Redis) -> bool:
        auto_release_time_ms = int(self.auto_release_time * 1000)
        extended = cast(Script, self._extend_script)(
            keys=(self.key,),
            args=(self._uuid, auto_release_time_ms),
            client=master,
        )
        return bool(extended)

    def __release_master(self, master: Redis) -> bool:
        released = cast(Script, self._release_script)(
            keys=(self.key,),
            args=(self._uuid,),
            client=master,
        )
        return bool(released)

    def __drift(self) -> float:
        return self.auto_release_time * self._CLOCK_DRIFT_FACTOR + .002

    def _acquire_masters(self,
                         *,
                         raise_on_redis_errors: bool | None = None,
                         ) -> bool:
        self._uuid = str(uuid.uuid4())
        self._extension_num = 0

        with ContextTimer() as timer, BailOutExecutor() as executor:
            futures = set()
            for master in self.masters:
                future = executor.submit(self.__acquire_master, master)
                futures.add(future)

            num_masters_acquired, redis_errors = 0, []
            for future in concurrent.futures.as_completed(futures):
                try:
                    num_masters_acquired += future.result()
                except RedisError as error:
                    redis_errors.append(error)
                    logger.exception(
                        '%s.__acquire_masters() caught %s',
                        self.__class__.__qualname__,
                        error.__class__.__qualname__,
                    )
                else:
                    if num_masters_acquired > len(self.masters) // 2:
                        validity_time = self.auto_release_time
                        validity_time -= self.__drift()
                        validity_time -= timer.elapsed() / 1000
                        if validity_time > 0:
                            return True

        with contextlib.suppress(ReleaseUnlockedLock):
            self.__release(raise_on_redis_errors=raise_on_redis_errors)
        self._check_enough_masters_up(raise_on_redis_errors, redis_errors)
        return False

    __acquire_masters = _acquire_masters

    def acquire(self,
                *,
                blocking: bool = True,
                timeout: float = -1,
                raise_on_redis_errors: bool | None = None,
                ) -> bool:
        '''Lock the lock.

        If blocking is True and timeout is -1, then wait for as long as
        necessary to acquire the lock.  Return True.

            >>> from redis import Redis
            >>> redis = Redis()
            >>> printer_lock_1 = Redlock(key='printer', masters={redis}, auto_release_time=.2)
            >>> printer_lock_1.acquire()
            True
            >>> timer = ContextTimer()
            >>> timer.start()
            >>> printer_lock_2 = Redlock(key='printer', masters={redis})
            >>> printer_lock_2.acquire()
            True
            >>> timer.elapsed() > printer_lock_1.auto_release_time * 1000
            True
            >>> printer_lock_2.release()

        If blocking is True and timeout is not -1, then wait for up to timeout
        seconds to acquire the lock.  Return True if the lock was acquired;
        False if it wasn't.

            >>> printer_lock_1.acquire()
            True
            >>> printer_lock_2.acquire(timeout=.5)
            True
            >>> printer_lock_2.release()

            >>> printer_lock_1.acquire()
            True
            >>> printer_lock_2.acquire(timeout=.1)
            False
            >>> with contextlib.suppress(ReleaseUnlockedLock):
            ...     printer_lock_1.release()

        If blocking is False and timeout is -1, then try just once right now to
        acquire the lock.  Return True if the lock was acquired; False if it
        wasn't.

            >>> printer_lock_1.acquire()
            True
            >>> printer_lock_2.acquire(blocking=False)
            False
            >>> printer_lock_1.release()
        '''
        acquire_masters = functools.partial(
            self.__acquire_masters,
            raise_on_redis_errors=raise_on_redis_errors,
        )

        if blocking:
            enqueued = False
            with ContextTimer() as timer:
                while timeout == -1 or timer.elapsed() / 1000 < timeout:
                    if acquire_masters():
                        if enqueued:
                            self.__log_time_enqueued(timer, acquired=True)
                        return True
                    enqueued = True
                    delay = random.uniform(0, self._RETRY_DELAY)  # nosec
                    time.sleep(delay)
            if enqueued:  # pragma: no cover
                self.__log_time_enqueued(timer, acquired=False)
            return False  # pragma: no cover

        if timeout == -1:
            return acquire_masters()

        raise ValueError("can't specify a timeout for a non-blocking call")

    __acquire = acquire

    def __log_time_enqueued(self, timer: ContextTimer, *, acquired: bool) -> None:
        key_suffix = self.key.split(':', maxsplit=1)[1]
        time_enqueued = math.ceil(timer.elapsed())
        logger.info(
            'source=pottery sample#redlock.enqueued.%s=%dms sample#redlock.acquired.%s=%d',
            key_suffix,
            time_enqueued,
            key_suffix,
            acquired,
        )

    def locked(self, *, raise_on_redis_errors: bool | None = None) -> float:
        '''How much longer we'll hold the lock (unless we extend or release it).

        If we don't currently hold the lock, then this method returns 0.

            >>> from redis import Redis
            >>> redis = Redis()
            >>> printer_lock_1 = Redlock(key='printer', masters={redis})
            >>> printer_lock_1.locked()
            0

            >>> printer_lock_2 = Redlock(key='printer', masters={redis})
            >>> printer_lock_2.acquire()
            True
            >>> printer_lock_1.locked()
            0
            >>> printer_lock_2.release()

        If we do currently hold the lock, then this method returns the current
        lease's Time To Live (TTL) in ms.

            >>> printer_lock_1.acquire()
            True
            >>> 9 < printer_lock_1.locked() < 10
            True
            >>> printer_lock_1.release()
        '''
        with ContextTimer() as timer, BailOutExecutor() as executor:
            futures = set()
            for master in self.masters:
                future = executor.submit(self.__acquired_master, master)
                futures.add(future)

            ttls, redis_errors = [], []
            for future in concurrent.futures.as_completed(futures):
                try:
                    ttl = future.result() / 1000
                except RedisError as error:
                    redis_errors.append(error)
                    logger.exception(
                        '%s.locked() caught %s',
                        self.__class__.__qualname__,
                        error.__class__.__qualname__,
                    )
                else:
                    if ttl:
                        ttls.append(ttl)
                        if len(ttls) > len(self.masters) // 2:  # pragma: no cover
                            validity_time = min(ttls)
                            validity_time -= self.__drift()
                            validity_time -= timer.elapsed() / 1000
                            return max(validity_time, 0)

        self._check_enough_masters_up(raise_on_redis_errors, redis_errors)
        return 0

    def extend(self, *, raise_on_redis_errors: bool | None = None) -> None:
        '''Extend our hold on the lock (if we currently hold it).

        Usage:

            >>> from redis import Redis
            >>> redis = Redis()
            >>> printer_lock = Redlock(key='printer', masters={redis})
            >>> printer_lock.acquire()
            True
            >>> 9 < printer_lock.locked() < 10
            True
            >>> time.sleep(1)
            >>> 8 < printer_lock.locked() < 9
            True
            >>> printer_lock.extend()
            >>> 9 < printer_lock.locked() < 10
            True
            >>> printer_lock.release()
        '''
        if self._extension_num >= self.num_extensions:
            raise TooManyExtensions(self.key, self.masters)

        with BailOutExecutor() as executor:
            futures = set()
            for master in self.masters:
                future = executor.submit(self.__extend_master, master)
                futures.add(future)

            num_masters_extended, redis_errors = 0, []
            for future in concurrent.futures.as_completed(futures):
                try:
                    num_masters_extended += future.result()
                except RedisError as error:
                    redis_errors.append(error)
                    logger.exception(
                        '%s.extend() caught %s',
                        self.__class__.__qualname__,
                        error.__class__.__qualname__,
                    )
                else:
                    if num_masters_extended > len(self.masters) // 2:
                        self._extension_num += 1
                        return

        self._check_enough_masters_up(raise_on_redis_errors, redis_errors)
        raise ExtendUnlockedLock(
            self.key,
            self.masters,
            redis_errors=redis_errors,
        )

    def release(self, *, raise_on_redis_errors: bool | None = None) -> None:
        '''Unlock the lock.

        Usage:

            >>> from redis import Redis
            >>> redis = Redis()
            >>> printer_lock = Redlock(key='printer', masters={redis})
            >>> bool(printer_lock.locked())
            False
            >>> printer_lock.acquire()
            True
            >>> bool(printer_lock.locked())
            True
            >>> printer_lock.release()
            >>> bool(printer_lock.locked())
            False
        '''
        with BailOutExecutor() as executor:
            futures = set()
            for master in self.masters:
                future = executor.submit(self.__release_master, master)
                futures.add(future)

            num_masters_released, redis_errors = 0, []
            for future in concurrent.futures.as_completed(futures):
                try:
                    num_masters_released += future.result()
                except RedisError as error:
                    redis_errors.append(error)
                    logger.exception(
                        '%s.release() caught %s',
                        self.__class__.__qualname__,
                        error.__class__.__qualname__,
                    )
                else:
                    if num_masters_released > len(self.masters) // 2:
                        return

        self._check_enough_masters_up(raise_on_redis_errors, redis_errors)
        raise ReleaseUnlockedLock(
            self.key,
            self.masters,
            redis_errors=redis_errors,
        )

    __release = release

    def __enter__(self) -> Redlock:
        '''You can use a Redlock as a context manager.

        Usage:

            >>> from redis import Redis
            >>> redis = Redis()
            >>> with Redlock(key='printer', masters={redis}) as printer_lock:
            ...     # Critical section - print stuff here.
            ...     print('printer_lock is locked')
            printer_lock is locked
            >>> bool(printer_lock.locked())
            False

            >>> with printer_lock:
            ...     # Critical section - print stuff here.
            ...     print('printer_lock is locked')
            printer_lock is locked
            >>> bool(printer_lock.locked())
            False
        '''
        acquired = self.__acquire(
            blocking=self.context_manager_blocking,
            timeout=self.context_manager_timeout,
        )
        if acquired:
            return self
        raise QuorumNotAchieved(self.key, self.masters)

    @overload
    def __exit__(self,
                 exc_type: None,
                 exc_value: None,
                 exc_traceback: None,
                 ) -> Literal[False]:
        raise NotImplementedError

    @overload
    def __exit__(self,
                 exc_type: Type[BaseException],
                 exc_value: BaseException,
                 exc_traceback: TracebackType,
                 ) -> Literal[False]:
        raise NotImplementedError

    def __exit__(self,
                 exc_type: Type[BaseException] | None,
                 exc_value: BaseException | None,
                 traceback: TracebackType | None,
                 ) -> Literal[False]:
        '''You can use a Redlock as a context manager.

        Usage:

            >>> from redis import Redis
            >>> redis = Redis()
            >>> with Redlock(key='printer', masters={redis}) as printer_lock:
            ...     # Critical section - print stuff here.
            ...     print('printer_lock is locked')
            printer_lock is locked
            >>> bool(printer_lock.locked())
            False

            >>> with printer_lock:
            ...     # Critical section - print stuff here.
            ...     print('printer_lock is locked')
            printer_lock is locked
            >>> bool(printer_lock.locked())
            False
        '''
        self.__release()
        return False

    def __repr__(self) -> str:
        return f'<{self.__class__.__qualname__} key={self.key}>'


def synchronize(*,
                key: str,
                masters: Iterable[Redis] = frozenset(),
                raise_on_redis_errors: bool = False,
                auto_release_time: float = Redlock._AUTO_RELEASE_TIME,
                blocking: bool = True,
                timeout: float = -1,
                ) -> Callable[[F], F]:
    '''Decorator to synchronize a function's execution across threads.

    synchronize() is a decorator that allows only one thread to execute a
    function at a time.  Under the hood, synchronize() uses a Redlock.  See
    help(Redlock) for more details.

    Keyword arguments:
        key -- a string that identifies your resource
        masters -- the Redis clients used to achieve quorum for this Redlock's
            state
        raise_on_redis_errors -- whether to raise the QuorumIsImplssible
            exception when too many Redis masters throw errors
        auto_release_time -- the timeout in seconds by which to automatically
            release this Redlock, unless it's already been released
        num_extensions -- the number of times that this Redlock's lease can be
            extended
        context_manager_blocking -- when using this Redlock as a context
            manager, whether to block when acquiring
        context_manager_timeout -- if context_manager_blocking, how long to wait
            when acquiring before giving up and raising the QuorumNotAchieved
            exception

    Usage:

        >>> @synchronize(key='synchronized-func', auto_release_time=1.5)
        ... def func():
        ...     # Only one thread can execute this function at a time.
        ...     return True
        ...
        >>> func()
        True
    '''
    RedlockFactory = functools.partial(
        Redlock,
        key=key,
        masters=masters,
        raise_on_redis_errors=raise_on_redis_errors,
        auto_release_time=auto_release_time,
        context_manager_blocking=blocking,
        context_manager_timeout=timeout,
    )

    def decorator(func: F) -> F:
        @functools.wraps(func)
        def wrapper(*args: Any, **kwargs: Any) -> Any:
            redlock = RedlockFactory()
            waiting_timer, holding_timer = ContextTimer(), ContextTimer()

            try:
                waiting_timer.start()
                with redlock:
                    waiting_timer.stop()
                    holding_timer.start()
                    return_value = func(*args, **kwargs)
                holding_timer.stop()
            finally:
                _log_synchronize(func, redlock, waiting_timer, holding_timer)

            return return_value
        return cast(F, wrapper)
    return decorator


def _log_synchronize(func: F,
                     redlock: Redlock,
                     waiting_timer: ContextTimer,
                     holding_timer: ContextTimer,
                     ) -> None:
    try:
        logger.info(
            '%s() waited for %s for %d ms; held for %d ms',
            func.__qualname__,
            redlock.key,
            waiting_timer.elapsed(),
            holding_timer.elapsed(),
        )
    except RuntimeError:
        # holding_timer.elapsed() threw a RuntimeError, which means that
        # holding_timer never started, which means that we never acquired the
        # lock / entered the critical section.
        logger.info(
            '%s() waited for %s for %d ms; never acquired lock',
            func.__qualname__,
            redlock.key,
            waiting_timer.elapsed(),
        )



# --------------------------------------------------------------------------- #
#   __init__.py                                                               #
#                                                                             #
#   Copyright © 2015-2022, Rajiv Bakulesh Shah, original author.              #
#                                                                             #
#   Licensed under the Apache License, Version 2.0 (the "License");           #
#   you may not use this file except in compliance with the License.          #
#   You may obtain a copy of the License at:                                  #
#       http://www.apache.org/licenses/LICENSE-2.0                            #
#                                                                             #
#   Unless required by applicable law or agreed to in writing, software       #
#   distributed under the License is distributed on an "AS IS" BASIS,         #
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  #
#   See the License for the specific language governing permissions and       #
#   limitations under the License.                                            #
# --------------------------------------------------------------------------- #
'''Redis for Humans.

Redis is awesome, but Redis commands are not always intuitive.  Pottery is a
Pythonic way to access Redis.  If you know how to use Python dicts, then you
already know how to use Pottery.  Pottery is useful for accessing Redis more
easily, and also for implementing microservice resilience patterns; and it has
been battle tested in production at scale.
'''


from typing import Tuple

# TODO: When we drop support for Python 3.7, change the following import to:
#   from typing import Final
try:
    from typing_extensions import Final
except ModuleNotFoundError:
    from typing import Final

__title__: Final[str] = 'tmp-pottery-test'
__version__: Final[str] = '0.0.1'
__description__: Final[str] = __doc__.split(sep='\n\n', maxsplit=1)[0]
__url__: Final[str] = 'https://github.com/brainix/pottery'
__author__: Final[str] = 'Rajiv Bakulesh Shah'
__author_email__: Final[str] = 'brainix@gmail.com'
__license__: Final[str] = 'Apache 2.0'
__keywords__: Final[str] = 'Redis client persistent storage'
__copyright__: Final[str] = f'Copyright © 2015-2022, {__author__}, original author.'


from .monkey import PotteryEncoder  # isort: skip

from .exceptions import PotteryError  # isort:skip
from .exceptions import KeyExistsError  # isort:skip
from .exceptions import RandomKeyError  # isort:skip
from .exceptions import QueueEmptyError  # isort:skip
from .exceptions import PrimitiveError  # isort:skip
from .exceptions import QuorumIsImpossible  # isort:skip
from .exceptions import QuorumNotAchieved  # isort:skip
from .exceptions import TooManyExtensions  # isort:skip
from .exceptions import ExtendUnlockedLock  # isort:skip
from .exceptions import ReleaseUnlockedLock  # isort:skip
from .exceptions import PotteryWarning  # isort:skip
from .exceptions import InefficientAccessWarning  # isort:skip

from .cache import CachedOrderedDict  # isort:skip
from .cache import redis_cache  # isort:skip
from .aionextid import AIONextID  # isort:skip
from .nextid import NextID  # isort:skip
from .aioredlock import AIORedlock  # isort:skip
from .redlock import Redlock  # isort:skip
from .redlock import synchronize  # isort:skip
from .timer import ContextTimer  # isort:skip

from .counter import RedisCounter
from .deque import RedisDeque
from .dict import RedisDict
from .list import RedisList
from .queue import RedisSimpleQueue
from .set import RedisSet


from .bloom import BloomFilter  # isort:skip
from .hyper import HyperLogLog  # isort:skip


__all__: Final[Tuple[str, ...]] = (
    'PotteryEncoder',

    'PotteryError',
    'KeyExistsError',
    'RandomKeyError',
    'QueueEmptyError',
    'PrimitiveError',
    'QuorumIsImpossible',
    'QuorumNotAchieved',
    'TooManyExtensions',
    'ExtendUnlockedLock',
    'ReleaseUnlockedLock',
    'PotteryWarning',
    'InefficientAccessWarning',

    'CachedOrderedDict',
    'redis_cache',
    'AIONextID',
    'NextID',
    'AIORedlock',
    'Redlock',
    'synchronize',
    'ContextTimer',

    'RedisCounter',
    'RedisDeque',
    'RedisDict',
    'RedisList',
    'RedisSimpleQueue',
    'RedisSet',
    'BloomFilter',
    'HyperLogLog',
)



# --------------------------------------------------------------------------- #
#   nextid.py                                                                 #
#                                                                             #
#   Copyright © 2015-2022, Rajiv Bakulesh Shah, original author.              #
#                                                                             #
#   Licensed under the Apache License, Version 2.0 (the "License");           #
#   you may not use this file except in compliance with the License.          #
#   You may obtain a copy of the License at:                                  #
#       http://www.apache.org/licenses/LICENSE-2.0                            #
#                                                                             #
#   Unless required by applicable law or agreed to in writing, software       #
#   distributed under the License is distributed on an "AS IS" BASIS,         #
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  #
#   See the License for the specific language governing permissions and       #
#   limitations under the License.                                            #
# --------------------------------------------------------------------------- #
'''Distributed Redis-powered monotonically increasing ID generator.

Rationale and algorithm description:
    http://antirez.com/news/102

Lua scripting:
    https://github.com/andymccurdy/redis-py#lua-scripting
'''


# TODO: Remove the following import after deferred evaluation of annotations
# because the default.
#   1. https://docs.python.org/3/whatsnew/3.7.html#whatsnew37-pep563
#   2. https://www.python.org/dev/peps/pep-0563/
#   3. https://www.python.org/dev/peps/pep-0649/
from __future__ import annotations

import concurrent.futures
import contextlib
from typing import ClassVar
from typing import Iterable
from typing import List
from typing import Tuple
from typing import Type
from typing import cast

from redis import Redis
from redis import RedisError
from redis.commands.core import Script

from .base import Primitive
from .base import logger
from .exceptions import QuorumIsImpossible
from .exceptions import QuorumNotAchieved
from .executor import BailOutExecutor


class Scripts:
    '''Parent class to define/register Lua scripts for Redis.

    Note that we only have to register these Lua scripts once -- so we do it on
    the first instantiation of NextID.
    '''

    __slots__: Tuple[str, ...] = tuple()

    _set_id_script: ClassVar[Script | None] = None

    def __init__(self,
                 *,
                 key: str = 'current',
                 masters: Iterable[Redis] = frozenset(),
                 raise_on_redis_errors: bool = False,
                 ) -> None:
        super().__init__(  # type: ignore
            key=key,
            masters=masters,
            raise_on_redis_errors=raise_on_redis_errors,
        )
        self.__register_set_id_script()

    # Preserve the Open-Closed Principle with name mangling.
    #   https://youtu.be/miGolgp9xq8?t=2086
    #   https://stackoverflow.com/a/38534939
    def __register_set_id_script(self) -> None:
        if self._set_id_script is None:
            class_name = self.__class__.__qualname__
            logger.info('Registering %s._set_id_script', class_name)
            master = next(iter(self.masters))  # type: ignore
            # Available since Redis 2.6.0:
            self.__class__._set_id_script = master.register_script('''
                local curr = tonumber(redis.call('get', KEYS[1]))
                local next = tonumber(ARGV[1])
                if curr == nil or curr < next then
                    redis.call('set', KEYS[1], next)
                    return next
                else
                    return nil
                end
            ''')


class NextID(Scripts, Primitive):
    '''Distributed Redis-powered monotonically increasing ID generator.

    This algorithm safely and reliably produces monotonically increasing IDs
    across threads, processes, and even machines, without a single point of
    failure.  Two caveats:

        1.  If many clients are generating IDs concurrently, then there may be
            "holes" in the sequence of IDs (e.g.: 1, 2, 6, 10, 11, 21, ...).

        2.  This algorithm scales to about 5,000 IDs per second (with 5 Redis
            masters).  If you need IDs faster than that, then you may want to
            consider other techniques.

    Rationale and algorithm description:
        http://antirez.com/news/102

    Clean up Redis for the doctest:

        >>> from redis import Redis
        >>> redis = Redis()
        >>> redis.delete('nextid:tweet-ids') in {0, 1}
        True

    Usage:

        >>> tweet_ids_1 = NextID(key='tweet-ids', masters={redis})
        >>> tweet_ids_2 = NextID(key='tweet-ids', masters={redis})
        >>> next(tweet_ids_1)
        1
        >>> next(tweet_ids_2)
        2
        >>> next(tweet_ids_1)
        3
        >>> tweet_ids_1.reset()
        >>> next(tweet_ids_1)
        1
    '''

    __slots__ = ('num_tries',)

    _KEY_PREFIX: ClassVar[str] = 'nextid'
    _NUM_TRIES: ClassVar[int] = 3

    def __init__(self,
                 *,
                 key: str = 'current',
                 masters: Iterable[Redis] = frozenset(),
                 raise_on_redis_errors: bool = False,
                 num_tries: int = _NUM_TRIES,
                 ) -> None:
        '''Initialize a NextID ID generator.

        Keyword arguments:
            key -- a string that identifies your ID sequence (e.g., 'tweets')
            masters -- the Redis clients used to achieve quorum for this ID
                generator
            raise_on_redis_errors -- whether to raise the QuorumIsImplssible
                exception when too many Redis masters throw errors
            num_tries -- the number of times to try to achieve quorum before
                giving up and raising the QuorumNotAchieved exception
        '''
        super().__init__(
            key=key,
            masters=masters,
            raise_on_redis_errors=raise_on_redis_errors,
        )
        self.num_tries = num_tries

    def __iter__(self) -> NextID:
        return self

    def __next__(self) -> int:
        suppressable_errors: List[Type[BaseException]] = [QuorumNotAchieved]
        if not self.raise_on_redis_errors:
            suppressable_errors.append(QuorumIsImpossible)
        for _ in range(self.num_tries):
            with contextlib.suppress(*suppressable_errors):
                next_id = self.__current_id + 1
                self.__current_id = next_id
                return next_id
        raise QuorumNotAchieved(self.key, self.masters)

    @property
    def __current_id(self) -> int:
        with BailOutExecutor() as executor:
            futures = set()
            for master in self.masters:
                future = executor.submit(master.get, self.key)
                futures.add(future)

            current_ids, redis_errors = [], []
            for future in concurrent.futures.as_completed(futures):
                try:
                    current_id = int(future.result() or b'0')
                except RedisError as error:
                    redis_errors.append(error)
                    logger.exception(
                        '%s.__current_id() getter caught %s',
                        self.__class__.__qualname__,
                        error.__class__.__qualname__,
                    )
                else:
                    current_ids.append(current_id)
                    if len(current_ids) > len(self.masters) // 2:  # pragma: no cover
                        return max(current_ids)

        self._check_enough_masters_up(None, redis_errors)
        raise QuorumNotAchieved(
            self.key,
            self.masters,
            redis_errors=redis_errors,
        )

    @__current_id.setter
    def __current_id(self, value: int) -> None:
        with BailOutExecutor() as executor:
            futures = set()
            for master in self.masters:
                future = executor.submit(
                    cast(Script, self._set_id_script),
                    keys=(self.key,),
                    args=(value,),
                    client=master,
                )
                futures.add(future)

            num_masters_set, redis_errors = 0, []
            for future in concurrent.futures.as_completed(futures):
                try:
                    num_masters_set += future.result() == value
                except RedisError as error:
                    redis_errors.append(error)
                    logger.exception(
                        '%s.__current_id() setter caught %s',
                        self.__class__.__qualname__,
                        error.__class__.__qualname__,
                    )
                else:
                    if num_masters_set > len(self.masters) // 2:  # pragma: no cover
                        return

        self._check_enough_masters_up(None, redis_errors)
        raise QuorumNotAchieved(
            self.key,
            self.masters,
            redis_errors=redis_errors,
        )

    def reset(self) -> None:
        'Reset the ID generator to 0.'
        with concurrent.futures.ThreadPoolExecutor() as executor:
            futures = set()
            for master in self.masters:
                future = executor.submit(master.delete, self.key)
                futures.add(future)

            num_masters_reset, redis_errors = 0, []
            for future in concurrent.futures.as_completed(futures):
                try:
                    future.result()
                except RedisError as error:
                    redis_errors.append(error)
                    logger.exception(
                        '%s.reset() caught %s',
                        self.__class__.__qualname__,
                        error.__class__.__qualname__,
                    )
                else:
                    num_masters_reset += 1
                    if num_masters_reset == len(self.masters):  # pragma: no cover
                        return

        self._check_enough_masters_up(None, redis_errors)
        raise QuorumNotAchieved(
            self.key,
            self.masters,
            redis_errors=redis_errors,
        )

    def __repr__(self) -> str:
        return f'<{self.__class__.__qualname__} key={self.key}>'



# --------------------------------------------------------------------------- #
#   hyper.py                                                                  #
#                                                                             #
#   Copyright © 2015-2022, Rajiv Bakulesh Shah, original author.              #
#                                                                             #
#   Licensed under the Apache License, Version 2.0 (the "License");           #
#   you may not use this file except in compliance with the License.          #
#   You may obtain a copy of the License at:                                  #
#       http://www.apache.org/licenses/LICENSE-2.0                            #
#                                                                             #
#   Unless required by applicable law or agreed to in writing, software       #
#   distributed under the License is distributed on an "AS IS" BASIS,         #
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  #
#   See the License for the specific language governing permissions and       #
#   limitations under the License.                                            #
# --------------------------------------------------------------------------- #


# TODO: Remove the following import after deferred evaluation of annotations
# because the default.
#   1. https://docs.python.org/3/whatsnew/3.7.html#whatsnew37-pep563
#   2. https://www.python.org/dev/peps/pep-0563/
#   3. https://www.python.org/dev/peps/pep-0649/
from __future__ import annotations

import uuid
import warnings
from typing import Generator
from typing import Iterable
from typing import List
from typing import cast

from redis import Redis

from .annotations import JSONTypes
from .annotations import RedisValues
from .base import Container
from .base import random_key
from .exceptions import InefficientAccessWarning


class HyperLogLog(Container):
    '''Redis-backed HyperLogLog with a Pythonic API.

    HyperLogLogs are an interesting data structure designed to answer the
    question, "How many distinct elements have I seen?"; but not the questions,
    "Have I seen this element before?" or "What are all of the elements that
    I've seen before?"  So think of HyperLogLogs as Python sets that you can add
    elements to and get the length of; but that you might not want to use to
    test element membership, and can't iterate through, or get elements out of.

    HyperLogLogs are probabilistic, which means that they&rsquo;re accurate
    within a margin of error up to 2%.  However, they can reasonably accurately
    estimate the cardinality (size) of vast datasets (like the number of unique
    Google searches issued in a day) with a tiny amount of storage (1.5 KB).

    Wikipedia article:
        https://en.wikipedia.org/wiki/HyperLogLog

    antirez's blog post:
        http://antirez.com/news/75

    Riak blog post:
        https://riak.com/posts/technical/what-in-the-hell-is-hyperloglog/index.html?p=13169.html

    Create a HyperLogLog and clean up Redis before the doctest:

        >>> google_searches = HyperLogLog(key='google-searches')
        >>> google_searches.clear()

    Insert an element into the HyperLogLog:

        >>> google_searches.add('sonic the hedgehog video game')

    See how many elements we've inserted into the HyperLogLog:

        >>> len(google_searches)
        1

    Insert multiple elements into the HyperLogLog:

        >>> google_searches.update({
        ...     'google in 1998',
        ...     'minesweeper',
        ...     'joey tribbiani',
        ...     'wizard of oz',
        ...     'rgb to hex',
        ...     'pac-man',
        ...     'breathing exercise',
        ...     'do a barrel roll',
        ...     'snake',
        ... })
        >>> len(google_searches)
        10

    Test for element membership in the HyperLogLog:

        >>> 'joey tribbiani' in google_searches
        True
        >>> 'jennifer aniston' in google_searches
        False

    Remove all of the elements from the HyperLogLog:

        >>> google_searches.clear()
    '''

    def __init__(self,
                 iterable: Iterable[RedisValues] = frozenset(),
                 *,
                 redis: Redis | None = None,
                 key: str = '',
                 ) -> None:
        '''Initialize the HyperLogLog.  O(n)

        Here, n is the number of elements in iterable that you want to insert
        into this HyperLogLog.
        '''
        super().__init__(redis=redis, key=key)
        self.__update(iterable)

    def add(self, value: RedisValues) -> None:
        'Add an element to the HyperLogLog.  O(1)'
        self.__update({value})

    def update(self, *objs: HyperLogLog | Iterable[RedisValues]) -> None:
        other_hll_keys: List[str] = []
        encoded_values: List[str] = []
        with self._watch(objs) as pipeline:
            for obj in objs:
                if isinstance(obj, self.__class__):
                    if not self._same_redis(obj):
                        raise RuntimeError(
                            f"can't update {self} with {obj} as they live on "
                            "different Redis instances/databases"
                        )
                    other_hll_keys.append(obj.key)
                else:
                    for value in cast(Iterable[JSONTypes], obj):
                        encoded_value = self._encode(value)
                        encoded_values.append(encoded_value)

            if other_hll_keys or len(encoded_values) > 1:
                warnings.warn(
                    cast(str, InefficientAccessWarning.__doc__),
                    InefficientAccessWarning,
                )

            pipeline.multi()  # Available since Redis 1.2.0
            pipeline.pfmerge(self.key, *other_hll_keys)  # Available since Redis 2.8.9
            pipeline.pfadd(self.key, *encoded_values)  # Available since Redis 2.8.9

    # Preserve the Open-Closed Principle with name mangling.
    #   https://youtu.be/miGolgp9xq8?t=2086
    #   https://stackoverflow.com/a/38534939
    __update = update

    def union(self,
              *objs: HyperLogLog | Iterable[RedisValues],
              redis: Redis | None = None,
              key: str = '',
              ) -> HyperLogLog:
        new_hll = self.__class__(redis=redis, key=key)
        new_hll.update(self, *objs)
        return new_hll

    def __len__(self) -> int:
        '''Return the approximate number of elements in the HyperLogLog.  O(1)

        Please note that this method returns an approximation, not an exact
        value, though it's quite accurate.
        '''
        return self.redis.pfcount(self.key)  # Available since Redis 2.8.9

    def __contains__(self, value: JSONTypes) -> bool:
        '''hll.__contains__(element) <==> element in hll.  O(1)

        Please note that this method *may* return false positives, but *never*
        returns false negatives.  This means that if `element in hll` evaluates
        to True, then you *may* have inserted the element into the HyperLogLog.
        But if `element in hll` evaluates to False, then you *must not* have
        inserted it.
        '''
        return next(self.__contains_many(value))

    def contains_many(self, *values: JSONTypes) -> Generator[bool, None, None]:
        '''Yield whether this HyperLogLog contains multiple elements.  O(n)

        Please note that this method *may* produce false positives, but *never*
        produces false negatives.  This means that if .contains_many() yields
        True, then you *may* have inserted the element into the HyperLogLog.
        But if .contains_many() yields False, then you *must not* have inserted
        it.
        '''
        if len(values) > 1:
            warnings.warn(
                cast(str, InefficientAccessWarning.__doc__),
                InefficientAccessWarning,
            )

        encoded_values = []
        for value in values:
            try:
                encoded_value = self._encode(value)
            except TypeError:
                encoded_value = str(uuid.uuid4())
            encoded_values.append(encoded_value)

        with self._watch() as pipeline:
            tmp_hll_key = random_key(redis=pipeline)
            pipeline.multi()  # Available since Redis 1.2.0
            for encoded_value in encoded_values:
                pipeline.copy(self.key, tmp_hll_key)  # Available since Redis 6.2.0
                pipeline.pfadd(tmp_hll_key, encoded_value)  # Available since Redis 2.8.9
                pipeline.unlink(tmp_hll_key)  # Available since Redis 4.0.0
            # Pluck out the results of the pipeline.pfadd() commands.  Ignore
            # the results of the enclosing pipeline.copy() and pipeline.unlink()
            # commands.
            cardinalities_changed = pipeline.execute()[1::3]  # Available since Redis 1.2.0

        for cardinality_changed in cardinalities_changed:
            yield not cardinality_changed

    __contains_many = contains_many

    def __repr__(self) -> str:
        'Return the string representation of the HyperLogLog.  O(1)'
        return f'<{self.__class__.__qualname__} key={self.key} len={len(self)}>'



# --------------------------------------------------------------------------- #
#   monkey.py                                                                 #
#                                                                             #
#   Copyright © 2015-2022, Rajiv Bakulesh Shah, original author.              #
#                                                                             #
#   Licensed under the Apache License, Version 2.0 (the "License");           #
#   you may not use this file except in compliance with the License.          #
#   You may obtain a copy of the License at:                                  #
#       http://www.apache.org/licenses/LICENSE-2.0                            #
#                                                                             #
#   Unless required by applicable law or agreed to in writing, software       #
#   distributed under the License is distributed on an "AS IS" BASIS,         #
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  #
#   See the License for the specific language governing permissions and       #
#   limitations under the License.                                            #
# --------------------------------------------------------------------------- #
'Monkey patches.'


# TODO: When we drop support for Python 3.9, remove the following import.  We
# only need it for X | Y union type annotations as of 2022-01-29.
from __future__ import annotations

import logging

# TODO: When we drop support for Python 3.7, change the following import to:
#   from typing import Final
try:
    from typing_extensions import Final
except ModuleNotFoundError:
    from typing import Final

logger: Final[logging.Logger] = logging.getLogger('pottery')
logger.addHandler(logging.NullHandler())


import functools  # isort: skip
import json  # isort: skip
from typing import Any  # isort: skip
from typing import Callable  # isort: skip

class PotteryEncoder(json.JSONEncoder):
    'Custom JSON encoder that can serialize Pottery containers.'

    def default(self, o: Any) -> Any:
        from pottery.base import Container
        if isinstance(o, Container):
            if hasattr(o, 'to_dict'):
                return o.to_dict()  # type: ignore
            if hasattr(o, 'to_list'):  # pragma: no cover
                return o.to_list()  # type: ignore
        return super().default(o)

def _decorate_dumps(func: Callable[..., str]) -> Callable[..., str]:
    'Decorate json.dumps() to use PotteryEncoder by default.'
    @functools.wraps(func)
    def wrapper(*args: Any,
                cls: type[json.JSONEncoder] = PotteryEncoder,
                **kwargs: Any,
                ) -> str:
        return func(*args, cls=cls, **kwargs)
    return wrapper

json.dumps = _decorate_dumps(json.dumps)

logger.info(
    'Monkey patched json.dumps() to be able to JSONify Pottery containers by '
    'default'
)



# --------------------------------------------------------------------------- #
#   bloom.py                                                                  #
#                                                                             #
#   Copyright © 2015-2022, Rajiv Bakulesh Shah, original author.              #
#                                                                             #
#   Licensed under the Apache License, Version 2.0 (the "License");           #
#   you may not use this file except in compliance with the License.          #
#   You may obtain a copy of the License at:                                  #
#       http://www.apache.org/licenses/LICENSE-2.0                            #
#                                                                             #
#   Unless required by applicable law or agreed to in writing, software       #
#   distributed under the License is distributed on an "AS IS" BASIS,         #
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  #
#   See the License for the specific language governing permissions and       #
#   limitations under the License.                                            #
# --------------------------------------------------------------------------- #


import abc
import functools
import itertools
import math
import uuid
import warnings
from typing import Any
from typing import Callable
from typing import Generator
from typing import Iterable
from typing import Set
from typing import cast

import mmh3
# TODO: When we drop support for Python 3.7, change the following import to:
#   from typing import final
from typing_extensions import final

from .annotations import F
from .annotations import JSONTypes
from .base import Container
from .exceptions import InefficientAccessWarning


# TODO: When we drop support for Python 3.7, stop using @_store_on_self().  Use
# @functools.cached_property() instead.
#   https://docs.python.org/3/library/functools.html#functools.cached_property
def _store_on_self(*, attr: str) -> Callable[[F], F]:
    "Decorator to store/cache a method's return value as an attribute on self."
    def decorator(func: F) -> F:
        @functools.wraps(func)
        def wrapper(self: Any, *args: Any, **kwargs: Any) -> Any:
            try:
                value = getattr(self, attr)
            except AttributeError:
                value = func(self, *args, **kwargs)
                setattr(self, attr, value)
            return value
        return cast(F, wrapper)
    return decorator


class BloomFilterABC(abc.ABC):
    '''Bloom filter abstract base class.

    This abstract base class:
        1. Defines the abstract methods that the concrete implementation class
           must define (having to do with encoding/decoding, I/O, and storage)
        2. Implements numerical recipes on the number of elements that you want
           to insert, and the acceptable false positive rate
    '''

    @abc.abstractmethod
    def _bit_offsets(self, value: JSONTypes) -> Generator[int, None, None]:
        for seed in range(self.num_hashes()):
            hash_ = mmh3.hash(cast(str, value), seed=seed)
            yield hash_ % self.size()

    @abc.abstractmethod
    def _num_bits_set(self) -> int:
        raise NotImplementedError

    @abc.abstractmethod
    def update(self, *iterables: Iterable[JSONTypes]) -> None:
        raise NotImplementedError

    @abc.abstractmethod
    def contains_many(self, *values: JSONTypes) -> Generator[bool, None, None]:
        raise NotImplementedError

    @final
    def __init__(self,
                 iterable: Iterable[JSONTypes] = frozenset(),
                 *args: Any,
                 num_elements: int,
                 false_positives: float,
                 **kwargs: Any,
                 ) -> None:
        '''Initialize the BloomFilter.  O(n * k)

        Here, n is the number of elements in iterable that you want to insert
        into this Bloom filter, and k is the number of times to run our hash
        functions on each element.
        '''
        super().__init__(*args, **kwargs)
        self.num_elements = num_elements
        self.false_positives = false_positives
        self.update(iterable)

    @final
    @_store_on_self(attr='_size')
    def size(self) -> int:
        '''The required number of bits (m) given n and p.

        This method returns the required number of bits (m) for the underlying
        string representing this Bloom filter given the the number of elements
        that you expect to insert (n) and your acceptable false positive
        probability (p).

        More about the formula that this method implements:
            https://en.wikipedia.org/wiki/Bloom_filter#Optimal_number_of_hash_functions
        '''
        size = (
            -self.num_elements
            * math.log(self.false_positives)
            / math.log(2)**2
        )
        size = math.ceil(size)
        return size

    @final
    @_store_on_self(attr='_num_hashes')
    def num_hashes(self) -> int:
        '''The number of hash functions (k) given m and n, minimizing p.

        This method returns the number of times (k) to run our hash functions
        on a given input string to compute bit offsets into the underlying
        string representing this Bloom filter.  m is the size in bits of the
        underlying string, n is the number of elements that you expect to
        insert, and p is your acceptable false positive probability.

        More about the formula that this method implements:
            https://en.wikipedia.org/wiki/Bloom_filter#Optimal_number_of_hash_functions
        '''
        num_hashes = self.size() / self.num_elements * math.log(2)
        num_hashes = math.ceil(num_hashes)
        return num_hashes

    @final
    def _bit_offsets_many(self,
                          *values: JSONTypes,
                          ) -> Generator[int, None, None]:
        for value in values:
            try:
                yield from self._bit_offsets(value)
            except TypeError:
                # value can't be encoded / converted to JSON.  Do a membership
                # test for a UUID in place of value.
                uuid4 = str(uuid.uuid4())
                yield from self._bit_offsets(uuid4)

    @final
    def add(self, value: JSONTypes) -> None:
        '''Add an element to the BloomFilter.  O(k)

        Here, k is the number of times to run our hash functions on a given
        input string to compute bit offests into the underlying string
        representing this Bloom filter.
        '''
        self.update({value})

    @final
    def __len__(self) -> int:
        '''Return the approximate the number of elements in the BloomFilter.  O(m)

        Here, m is the size in bits of the underlying string representing this
        Bloom filter.

        Please note that this method returns an approximation, not an exact
        value, though it's quite accurate.  The accuracy can be tuned with the
        num_elements and false_positives arguments to .__init__(), at the
        expense of storage and CPU time.

        More about the formula that this method implements:
            https://en.wikipedia.org/wiki/Bloom_filter#Approximating_the_number_of_items_in_a_Bloom_filter
        '''
        len_ = (
            -self.size()
            / self.num_hashes()
            * math.log(1 - self._num_bits_set() / self.size())
        )
        return math.floor(len_)

    @final
    def __contains__(self, value: JSONTypes) -> bool:
        '''bf.__contains__(element) <==> element in bf.  O(k)

        Here, k is the number of times to run our hash functions on a given
        input string to compute bit offests into the underlying string
        representing this Bloom filter.

        Please note that this method *may* return false positives, but *never*
        returns false negatives.  This means that if `element in bf` evaluates
        to True, then you *may* have inserted the element into the Bloom filter.
        But if `element in bf` evaluates to False, then you *must not* have
        inserted it.
        '''
        return next(self.contains_many(value))


class BloomFilter(BloomFilterABC, Container):
    '''Redis-backed Bloom filter with an API similar to Python sets.

    Bloom filters are a powerful data structure that help you to answer the
    question, "Have I seen this element before?" but not the question, "What
    are all of the elements that I've seen before?"  So think of Bloom filters
    as Python sets that you can add elements to and use to test element
    membership, but that you can't iterate through or get elements back out of.

    Bloom filters are probabilistic, which means that they can sometimes
    generate false positives (as in, they may report that you've seen a
    particular element before even though you haven't).  But they will never
    generate false negatives (so every time that they report that you haven't
    seen a particular element before, you really must never have seen it).  You
    can tune your acceptable false positive probability, though at the expense
    of the storage size and the element insertion/lookup time of your Bloom
    filter.

    Wikipedia article:
        https://en.wikipedia.org/wiki/Bloom_filter

    Reference implementation:
        http://www.maxburstein.com/blog/creating-a-simple-bloom-filter/

    Instantiate a Bloom filter and clean up Redis before the doctest:

        >>> dilberts = BloomFilter(
        ...     num_elements=100,
        ...     false_positives=0.01,
        ...     key='dilberts',
        ... )
        >>> dilberts.clear()

    Here, num_elements represents the number of elements that you expect to
    insert into your BloomFilter, and false_positives represents your
    acceptable false positive probability.  Using these two parameters,
    BloomFilter automatically computes its own storage size and number of times
    to run its hash functions on element insertion/lookup such that it can
    guarantee a false positive rate at or below what you can tolerate, given
    that you're going to insert your specified number of elements.

    Insert an element into the Bloom filter:

        >>> dilberts.add('rajiv')

    Test for element membership in the Bloom filter:

        >>> 'rajiv' in dilberts
        True
        >>> 'raj' in dilberts
        False
        >>> 'dan' in dilberts
        False

    See how many elements we've inserted into the Bloom filter:

        >>> len(dilberts)
        1

    Note that BloomFilter.__len__() is an approximation, not an exact value,
    though it's quite accurate.

    Insert multiple elements into the Bloom filter:

        >>> dilberts.update({'raj', 'dan'})
        >>> len(dilberts)
        3

    Remove all of the elements from the Bloom filter:

        >>> dilberts.clear()
    '''

    def _bit_offsets(self, value: JSONTypes) -> Generator[int, None, None]:
        '''The bit offsets to set/check in this Bloom filter for a given value.

        Instantiate a Bloom filter:

            >>> dilberts = BloomFilter(
            ...     num_elements=100,
            ...     false_positives=0.01,
            ...     key='dilberts',
            ... )

        Now let's look at a few examples:

            >>> tuple(dilberts._bit_offsets('rajiv'))
            (183, 319, 787, 585, 8, 471, 711)
            >>> tuple(dilberts._bit_offsets('raj'))
            (482, 875, 725, 667, 109, 714, 595)
            >>> tuple(dilberts._bit_offsets('dan'))
            (687, 925, 954, 707, 615, 914, 620)

        Thus, if we want to insert the value 'rajiv' into our Bloom filter,
        then we must set bits 183, 319, 787, 585, 8, 471, and 711 all to 1.  If
        any/all of them are already 1, no problems.

        Similarly, if we want to check to see if the value 'rajiv' is in our
        Bloom filter, then we must check to see if the bits 183, 319, 787, 585,
        8, 471, and 711 are all set to 1.  If even one of those bits is set to
        0, then the value 'rajiv' must never have been inserted into our Bloom
        filter.  But if all of those bits are set to 1, then the value 'rajiv'
        has *probably* been inserted into our Bloom filter.
        '''
        encoded_value = self._encode(value)
        return super()._bit_offsets(encoded_value)

    def _num_bits_set(self) -> int:
        '''The number of bits set to 1 in this Bloom filter.  O(m)

        Here, m is the size in bits of the underlying string representing this
        Bloom filter.
        '''
        return self.redis.bitcount(self.key)  # Available since Redis 2.6.0

    def update(self, *iterables: Iterable[JSONTypes]) -> None:
        '''Populate the Bloom filter with the elements in iterables.  O(n * k)

        Here, n is the number of elements in iterables that you want to insert
        into this Bloom filter, and k is the number of times to run our hash
        functions on each element.
        '''
        with self._watch(*iterables) as pipeline:
            values = tuple(itertools.chain(*iterables))
            if len(values) > 1:
                warnings.warn(
                    cast(str, InefficientAccessWarning.__doc__),
                    InefficientAccessWarning,
                )

            bit_offsets: Set[int] = set()
            for value in values:
                bit_offsets.update(self._bit_offsets(value))

            pipeline.multi()  # Available since Redis 1.2.0
            for bit_offset in bit_offsets:
                pipeline.setbit(self.key, bit_offset, 1)  # Available since Redis 2.2.0

    def contains_many(self, *values: JSONTypes) -> Generator[bool, None, None]:
        '''Yield whether this Bloom filter contains multiple elements.  O(n)

        Please note that this method *may* produce false positives, but *never*
        produces false negatives.  This means that if .contains_many() yields
        True, then you *may* have inserted the element into the Bloom filter.
        But if .contains_many() yields False, then you *must not* have inserted
        it.
        '''
        if len(values) > 1:
            warnings.warn(
                cast(str, InefficientAccessWarning.__doc__),
                InefficientAccessWarning,
            )

        with self._watch() as pipeline:
            pipeline.multi()  # Available since Redis 1.2.0
            for bit_offset in self._bit_offsets_many(*values):
                pipeline.getbit(self.key, bit_offset)  # Available since Redis 2.2.0
            bits = iter(pipeline.execute())  # Available since Redis 1.2.0

        # I stole this recipe from here:
        #   https://stackoverflow.com/a/61435714
        #
        # TODO: When we drop support for Python 3.7, rewrite the following loop
        # using the walrus operator, like in the Stack Overflow answer linked
        # above.
        bits_per_chunk = self.num_hashes()
        while True:
            bits_in_chunk = tuple(itertools.islice(bits, bits_per_chunk))
            if not bits_in_chunk:
                break
            yield all(bits_in_chunk)

    def __repr__(self) -> str:
        'Return the string representation of the BloomFilter.  O(1)'
        return f'<{self.__class__.__qualname__} key={self.key}>'



# --------------------------------------------------------------------------- #
#   cache.py                                                                  #
#                                                                             #
#   Copyright © 2015-2022, Rajiv Bakulesh Shah, original author.              #
#                                                                             #
#   Licensed under the Apache License, Version 2.0 (the "License");           #
#   you may not use this file except in compliance with the License.          #
#   You may obtain a copy of the License at:                                  #
#       http://www.apache.org/licenses/LICENSE-2.0                            #
#                                                                             #
#   Unless required by applicable law or agreed to in writing, software       #
#   distributed under the License is distributed on an "AS IS" BASIS,         #
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  #
#   See the License for the specific language governing permissions and       #
#   limitations under the License.                                            #
# --------------------------------------------------------------------------- #


# TODO: When we drop support for Python 3.9, remove the following import.  We
# only need it for X | Y union type annotations as of 2022-01-29.
from __future__ import annotations

import collections
import functools
import itertools
from typing import Any
from typing import Callable
from typing import ClassVar
from typing import Collection
from typing import Hashable
from typing import Iterable
from typing import Mapping
from typing import NamedTuple
from typing import Tuple
from typing import TypeVar
from typing import Union
from typing import cast

from redis import Redis
from redis.exceptions import WatchError
# TODO: When we drop support for Python 3.7, change the following import to:
#   from typing import Final
from typing_extensions import Final

from .annotations import JSONTypes
from .base import _default_redis
from .base import logger
from .base import random_key
from .dict import RedisDict


F = TypeVar('F', bound=Callable[..., JSONTypes])

UpdateMap = Mapping[JSONTypes, Union[JSONTypes, object]]
UpdateItem = Tuple[JSONTypes, Union[JSONTypes, object]]
UpdateIter = Iterable[UpdateItem]
UpdateArg = Union[UpdateMap, UpdateIter]

_DEFAULT_TIMEOUT: Final[int] = 60   # seconds


class CacheInfo(NamedTuple):
    '''Caching decorator information.

    This CacheInfo named tuple is compatible with the one in functools:
        https://github.com/python/cpython/blob/7a34380ad788886f5ad50d4175ceb2d5715b8cff/Lib/functools.py#L430
    '''
    hits: int = 0
    misses: int = 0
    maxsize: int | None = None
    currsize: int = 0


def _arg_hash(*args: Hashable, **kwargs: Hashable) -> int:
    kwargs_items = frozenset(kwargs.items())
    return hash((args, kwargs_items))


def redis_cache(*,  # NoQA: C901
                redis: Redis | None = None,
                key: str | None = None,
                timeout: int | None = _DEFAULT_TIMEOUT,
                ) -> Callable[[F], F]:
    '''Redis-backed caching decorator with an API like functools.lru_cache().

    Arguments to the original underlying function must be hashable, and return
    values from the function must be JSON serializable.

    Additionally, this decorator provides the following functions:

    f.__wrapped__(*args, **kwargs)
        Access the original underlying function.  This is useful for
        introspection, for bypassing the cache, or for rewrapping the function
        with a different cache.

    f.__bypass__(*args, **kwargs)
        Force a cache reset for your args/kwargs.  Bypass the cache lookup,
        call the original underlying function, then cache the results for
        future calls to f(*args, **kwargs).

    f.cache_info()
        Return a NamedTuple showing hits, misses, maxsize, and currsize.  This
        information is helpful for measuring the effectiveness of the cache.

        Note that maxsize is always None, meaning that this cache is always
        unbounded.  maxsize is only included for compatibility with
        functools.lru_cache().

        While redis_cache() is thread-safe, also note that hits/misses only
        instrument your local process - not other processes, even if connected
        to the same Redis-backed redis_cache() key.  And in some cases,
        hits/misses may be incorrect in multiprocess/distributed applications.

        That said, currsize is always correct, even if other remote processes
        modify the same Redis-backed redis_cache() key.

    f.cache_clear()
        Clear/invalidate the entire cache (for all args/kwargs previously
        cached) for your function.

    In general, you should only use redis_cache() when you want to reuse
    previously computed values.  Accordingly, it doesn't make sense to cache
    functions with side-effects or impure functions such as time() or random().

    However, unlike functools.lru_cache(), redis_cache() reconstructs
    previously cached objects on each cache hit.  Therefore, you can use
    redis_cache() for a function that needs to create a distinct mutable object
    on each call.
    '''

    if redis is None:
        redis = _default_redis

    def decorator(func: F) -> F:
        nonlocal redis, key
        if key is None:
            key = random_key(redis=cast(Redis, redis))
            logger.warning(
                "Self-assigning key redis_cache(key='%s') for function %s",
                key,
                func.__qualname__,
            )
        cache = RedisDict(redis=redis, key=key)
        hits, misses = 0, 0

        @functools.wraps(func)
        def wrapper(*args: Hashable, **kwargs: Hashable) -> JSONTypes:
            nonlocal hits, misses
            hash_ = _arg_hash(*args, **kwargs)
            try:
                return_value = cache[hash_]
                hits += 1
            except KeyError:
                return_value = func(*args, **kwargs)
                cache[hash_] = return_value
                misses += 1
            if timeout:
                cast(Redis, redis).expire(cast(str, key), timeout)
            return return_value

        @functools.wraps(func)
        def bypass(*args: Hashable, **kwargs: Hashable) -> JSONTypes:
            hash_ = _arg_hash(*args, **kwargs)
            return_value = func(*args, **kwargs)
            cache[hash_] = return_value
            if timeout:
                cast(Redis, redis).expire(cast(str, key), timeout)
            return return_value

        def cache_info() -> CacheInfo:
            return CacheInfo(
                hits=hits,
                misses=misses,
                maxsize=None,
                currsize=len(cache),
            )

        def cache_clear() -> None:
            nonlocal hits, misses
            cast(Redis, redis).unlink(cast(str, key))
            hits, misses = 0, 0

        wrapper.__wrapped__ = func  # type: ignore
        wrapper.__bypass__ = bypass  # type: ignore
        wrapper.cache_info = cache_info  # type: ignore
        wrapper.cache_clear = cache_clear  # type: ignore
        return cast(F, wrapper)
    return decorator


def _set_expiration(func: F) -> F:
    @functools.wraps(func)
    def wrapper(self: Any, *args: Any, **kwargs: Any) -> Any:
        value = func(self, *args, **kwargs)
        if self._timeout:
            self._cache.redis.expire(self._cache.key, self._timeout)  # Available since Redis 1.0.0
        return value
    return cast(F, wrapper)


class CachedOrderedDict(collections.OrderedDict):
    '''Redis-backed container that extends Python's OrderedDicts.

    The best way that I can explain CachedOrderedDict is through an example
    use-case.  Imagine that your search engine returns document IDs, which then
    you have to hydrate into full documents via the database to return to the
    client.  The data structure used to represent such search results must have
    the following properties:

        1. It must preserve the order of the document IDs returned by the
           search engine
        2. It must map document IDs to hydrated documents
        3. It must cache previously hydrated documents

    Properties 1 and 2 are satisfied by Python's collections.OrderedDict.
    However, CachedOrderedDict extends Python's OrderedDict to also satisfy
    property 3.
    '''

    _SENTINEL: ClassVar[object] = object()
    _NUM_TRIES: ClassVar[int] = 3

    @_set_expiration
    def __init__(self,
                 *,
                 redis_client: Redis | None = None,
                 redis_key: str | None = None,
                 dict_keys: Iterable[JSONTypes] = tuple(),
                 num_tries: int = _NUM_TRIES,
                 timeout: int | None = _DEFAULT_TIMEOUT,
                 ) -> None:
        self._num_tries = num_tries
        self._timeout = timeout
        init_cache = functools.partial(
            RedisDict,
            redis=redis_client,
            key=redis_key,
        )
        self._cache = self.__retry(init_cache)
        self._misses = set()

        # We have to iterate over dict_keys multiple times, so cast it to a
        # tuple.  This allows the caller to pass in a generator for dict_keys,
        # and we can still iterate over it multiple times.
        items, dict_keys = [], tuple(dict_keys)
        if dict_keys:
            encoded_keys = (
                self._cache._encode(dict_key) for dict_key in dict_keys
            )
            encoded_values = self._cache.redis.hmget(  # Available since Redis 2.0.0
                self._cache.key,
                *encoded_keys,
            )
            for dict_key, encoded_value in zip(dict_keys, encoded_values):
                if encoded_value is None:
                    self._misses.add(dict_key)
                    value = self._SENTINEL
                else:
                    value = self._cache._decode(encoded_value)
                item = (dict_key, value)
                items.append(item)
        super().__init__()
        self.__update(items)

    def misses(self) -> Collection[JSONTypes]:
        return frozenset(self._misses)

    @_set_expiration
    def __setitem__(self,
                    dict_key: JSONTypes,
                    value: JSONTypes | object,
                    ) -> None:
        'Set self[dict_key] to value.'
        if value is not self._SENTINEL:
            self._cache[dict_key] = value
            self._misses.discard(dict_key)
        super().__setitem__(dict_key, value)

    @_set_expiration
    def setdefault(self,
                   dict_key: JSONTypes,
                   default: JSONTypes = None,
                   ) -> JSONTypes:
        '''Insert key with a value of default if key is not in the dictionary.

        Return the value for key if key is in the dictionary, else default.
        '''
        retriable_setdefault = functools.partial(
            self.__retriable_setdefault,
            dict_key,
            default=default,
        )
        self.__retry(retriable_setdefault)
        self._misses.discard(dict_key)
        if dict_key not in self or self[dict_key] is self._SENTINEL:
            value = self[dict_key] = default
        else:  # pragma: no cover
            value = self[dict_key]
        return value

    # Preserve the Open-Closed Principle with name mangling.
    #   https://youtu.be/miGolgp9xq8?t=2086
    #   https://stackoverflow.com/a/38534939
    def __retriable_setdefault(self,
                               dict_key: JSONTypes,
                               default: JSONTypes = None,
                               ) -> None:
        with self._cache._watch() as pipeline:
            if dict_key not in self._cache:  # pragma: no cover
                pipeline.multi()  # Available since Redis 1.2.0
                # The following line is equivalent to: self._cache[dict_key] = default
                pipeline.hset(  # Available since Redis 2.0.0
                    self._cache.key,
                    self._cache._encode(dict_key),
                    self._cache._encode(default),
                )

    def __retry(self, callable: Callable[[], Any], *, try_num: int = 0) -> Any:
        try:
            return callable()
        except WatchError:  # pragma: no cover
            if try_num < self._num_tries - 1:
                return self.__retry(callable, try_num=try_num+1)
            raise

    @_set_expiration
    def update(self,  # type: ignore
               arg: UpdateArg = tuple(),
               **kwargs: JSONTypes | object,
               ) -> None:
        '''D.update([E, ]**F) -> None.  Update D from dict/iterable E and F.
        If E is present and has an .items() method, then does:  for k in E: D[k] = E[k]
        If E is present and lacks an .items() method, then does:  for k, v in E: D[k] = v
        In either case, this is followed by: for k in F:  D[k] = F[k]

        The base class, OrderedDict, has an .update() method that works just
        fine.  The trouble is that it executes multiple calls to
        self.__setitem__() therefore multiple round trips to Redis.  This
        overridden .update() makes a single bulk call to Redis.
        '''
        to_cache = {}
        if isinstance(arg, collections.abc.Mapping):
            arg = arg.items()
        items = itertools.chain(arg, kwargs.items())
        for dict_key, value in items:
            if value is not self._SENTINEL:
                to_cache[dict_key] = value
                self._misses.discard(cast(JSONTypes, dict_key))
            super().__setitem__(dict_key, value)
        self._cache.update(to_cache)

    __update = update



# --------------------------------------------------------------------------- #
#   setup.py                                                                  #
#                                                                             #
#   Copyright © 2015-2022, Rajiv Bakulesh Shah, original author.              #
#                                                                             #
#   Licensed under the Apache License, Version 2.0 (the "License");           #
#   you may not use this file except in compliance with the License.          #
#   You may obtain a copy of the License at:                                  #
#       http://www.apache.org/licenses/LICENSE-2.0                            #
#                                                                             #
#   Unless required by applicable law or agreed to in writing, software       #
#   distributed under the License is distributed on an "AS IS" BASIS,         #
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  #
#   See the License for the specific language governing permissions and       #
#   limitations under the License.                                            #
# --------------------------------------------------------------------------- #


import pathlib

from setuptools import find_packages
from setuptools import setup

import pottery


_package_dir = pathlib.Path(__file__).parent
_long_description = (_package_dir / 'README.md').read_text()


setup(
    name=pottery.__title__,
    version=pottery.__version__,
    description=pottery.__description__,
    long_description=_long_description,
    long_description_content_type='text/markdown',
    url=pottery.__url__,
    author=pottery.__author__,
    author_email=pottery.__author_email__,
    license=pottery.__license__,
    classifiers=[
        'License :: OSI Approved :: Apache Software License',
        'Intended Audience :: Developers',
        'Natural Language :: English',
        'Development Status :: 4 - Beta',
        'Topic :: Database :: Front-Ends',
        'Topic :: System :: Distributed Computing',
        'Topic :: Utilities',
        'Programming Language :: Python :: 3 :: Only',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Framework :: AsyncIO',
        'Typing :: Typed',
    ],
    keywords=pottery.__keywords__,
    python_requires='>=3.7, <4',
    install_requires=('redis>=4.2.0rc1, <5', 'mmh3', 'typing_extensions'),
    extras_require={},
    packages=find_packages(exclude=('contrib', 'docs', 'tests*')),
    package_data={'pottery': ('py.typed',)},
    data_files=tuple(),
    entry_points={},
)



# --------------------------------------------------------------------------- #
#   exceptions.py                                                             #
#                                                                             #
#   Copyright © 2015-2022, Rajiv Bakulesh Shah, original author.              #
#                                                                             #
#   Licensed under the Apache License, Version 2.0 (the "License");           #
#   you may not use this file except in compliance with the License.          #
#   You may obtain a copy of the License at:                                  #
#       http://www.apache.org/licenses/LICENSE-2.0                            #
#                                                                             #
#   Unless required by applicable law or agreed to in writing, software       #
#   distributed under the License is distributed on an "AS IS" BASIS,         #
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  #
#   See the License for the specific language governing permissions and       #
#   limitations under the License.                                            #
# --------------------------------------------------------------------------- #


# TODO: When we drop support for Python 3.9, remove the following import.  We
# only need it for X | Y union type annotations as of 2022-01-29.
from __future__ import annotations

from dataclasses import dataclass
from queue import Empty
from typing import Iterable

from redis import Redis
from redis import RedisError


@dataclass
class PotteryError(Exception):
    'Base exception class for Pottery containers.'

    redis: Redis
    key: str | None = None

class KeyExistsError(PotteryError):
    'Initializing a container on a Redis key that already exists.'

class RandomKeyError(PotteryError, RuntimeError):
    "Can't create a random Redis key; all of our attempts already exist."

class QueueEmptyError(PotteryError, Empty):
    'Non-blocking .get() or .get_nowait() called on RedisQueue which is empty.'


@dataclass
class PrimitiveError(Exception):
    'Base exception class for distributed primitives.'

    key: str
    masters: Iterable[Redis]
    redis_errors: Iterable[RedisError] = tuple()

class QuorumNotAchieved(PrimitiveError, RuntimeError):
    'Consensus-based algorithm could not achieve quorum.'

class TooManyExtensions(PrimitiveError, RuntimeError):
    'Redlock has been extended too many times.'

class ExtendUnlockedLock(PrimitiveError, RuntimeError):
    'Attempting to extend an unlocked Redlock.'

class ReleaseUnlockedLock(PrimitiveError, RuntimeError):
    'Attempting to release an unlocked Redlock.'

class QuorumIsImpossible(PrimitiveError, RuntimeError):
    'Too many Redis masters threw RedisErrors; quorum can not be achieved.'


class PotteryWarning(Warning):
    'Base warning class for Pottery containers.'

class InefficientAccessWarning(PotteryWarning):
    'Doing an O(n) Redis operation.'



# --------------------------------------------------------------------------- #
#   aioredlock.py                                                             #
#                                                                             #
#   Copyright © 2015-2022, Rajiv Bakulesh Shah, original author.              #
#                                                                             #
#   Licensed under the Apache License, Version 2.0 (the "License");           #
#   you may not use this file except in compliance with the License.          #
#   You may obtain a copy of the License at:                                  #
#       http://www.apache.org/licenses/LICENSE-2.0                            #
#                                                                             #
#   Unless required by applicable law or agreed to in writing, software       #
#   distributed under the License is distributed on an "AS IS" BASIS,         #
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  #
#   See the License for the specific language governing permissions and       #
#   limitations under the License.                                            #
# --------------------------------------------------------------------------- #
'''Asynchronous distributed Redis-powered lock.

This algorithm safely and reliably provides a mutually-exclusive locking
primitive to protect a resource shared across coroutines, threads, processes,
and even machines, without a single point of failure.

Rationale and algorithm description:
    http://redis.io/topics/distlock

Reference implementations:
    https://github.com/antirez/redlock-rb
    https://github.com/SPSCommerce/redlock-py

Lua scripting:
    https://github.com/andymccurdy/redis-py#lua-scripting
'''


# TODO: Remove the following import after deferred evaluation of annotations
# because the default.
#   1. https://docs.python.org/3/whatsnew/3.7.html#whatsnew37-pep563
#   2. https://www.python.org/dev/peps/pep-0563/
#   3. https://www.python.org/dev/peps/pep-0649/
from __future__ import annotations

import asyncio
import contextlib
import functools
import math
import random
import uuid
from types import TracebackType
from typing import ClassVar
from typing import Iterable
from typing import Type

from redis import RedisError
from redis.asyncio import Redis as AIORedis  # type: ignore

from .base import AIOPrimitive
from .base import logger
from .exceptions import ExtendUnlockedLock
from .exceptions import QuorumNotAchieved
from .exceptions import ReleaseUnlockedLock
from .exceptions import TooManyExtensions
from .redlock import Redlock
from .redlock import Scripts
from .timer import ContextTimer


class AIORedlock(Scripts, AIOPrimitive):
    '''Asynchronous distributed Redis-powered lock.

    This algorithm safely and reliably provides a mutually-exclusive locking
    primitive to protect a resource shared across coroutines, threads,
    processes, and even machines, without a single point of failure.

    Rationale and algorithm description:
        http://redis.io/topics/distlock

    Usage:

        >>> import asyncio
        >>> from redis.asyncio import Redis as AIORedis
        >>> async def main():
        ...     aioredis = AIORedis.from_url('redis://localhost:6379/1')
        ...     shower = AIORedlock(key='shower', masters={aioredis})
        ...     if await shower.acquire():
        ...         # Critical section - no other coroutine can enter while we hold the lock.
        ...         print(f"shower is {'occupied' if await shower.locked() else 'available'}")
        ...         await shower.release()
        ...     print(f"shower is {'occupied' if await shower.locked() else 'available'}")
        >>> asyncio.run(main(), debug=True)
        shower is occupied
        shower is available

    AIORedlocks time out (by default, after 10 seconds).  You should take care
    to ensure that your critical section completes well within the timeout.  The
    reasons that AIORedlocks time out are to preserve "liveness"
    (http://redis.io/topics/distlock#liveness-arguments) and to avoid deadlocks
    (in the event that a process dies inside a critical section before it
    releases its lock).

        >>> async def main():
        ...     aioredis = AIORedis.from_url('redis://localhost:6379/1')
        ...     shower = AIORedlock(key='shower', masters={aioredis}, auto_release_time=.2)
        ...     if await shower.acquire():
        ...         # Critical section - no other coroutine can enter while we hold the lock.
        ...         await asyncio.sleep(shower.auto_release_time)
        ...     print(f"shower is {'occupied' if await shower.locked() else 'available'}")
        >>> asyncio.run(main(), debug=True)
        shower is available

    If 10 seconds isn't enough to complete executing your critical section,
    then you can specify your own timeout:

        >>> async def main():
        ...     aioredis = AIORedis.from_url('redis://localhost:6379/1')
        ...     shower = AIORedlock(key='shower', masters={aioredis}, auto_release_time=.2)
        ...     if await shower.acquire():
        ...         # Critical section - no other coroutine can enter while we hold the lock.
        ...         await asyncio.sleep(shower.auto_release_time / 2)
        ...     print(f"shower is {'occupied' if await shower.locked() else 'available'}")
        ...     await asyncio.sleep(shower.auto_release_time / 2)
        ...     print(f"shower is {'occupied' if await shower.locked() else 'available'}")
        >>> asyncio.run(main(), debug=True)
        shower is occupied
        shower is available

    You can use an AIORedlock as a context manager:

        >>> async def main():
        ...     aioredis = AIORedis.from_url('redis://localhost:6379/1')
        ...     shower = AIORedlock(key='shower', masters={aioredis})
        ...     async with shower:
        ...         print(f"shower is {'occupied' if await shower.locked() else 'available'}")
        ...         # Critical section - no other coroutine can enter while we hold the lock.
        ...     print(f"shower is {'occupied' if await shower.locked() else 'available'}")
        >>> asyncio.run(main(), debug=True)
        shower is occupied
        shower is available
    '''

    __slots__ = (
        'auto_release_time',
        'num_extensions',
        'context_manager_blocking',
        'context_manager_timeout',
        '_uuid',
        '_extension_num',
    )

    _KEY_PREFIX: ClassVar[str] = Redlock._KEY_PREFIX
    _AUTO_RELEASE_TIME: ClassVar[float] = Redlock._AUTO_RELEASE_TIME
    _CLOCK_DRIFT_FACTOR: ClassVar[float] = Redlock._CLOCK_DRIFT_FACTOR
    _RETRY_DELAY: ClassVar[float] = Redlock._RETRY_DELAY
    _NUM_EXTENSIONS: ClassVar[int] = Redlock._NUM_EXTENSIONS

    def __init__(self,  # type: ignore
                 *,
                 key: str,
                 masters: Iterable[AIORedis],
                 raise_on_redis_errors: bool = False,
                 auto_release_time: float = _AUTO_RELEASE_TIME,
                 num_extensions: int = _NUM_EXTENSIONS,
                 context_manager_blocking: bool = True,
                 context_manager_timeout: float = -1,
                 ) -> None:
        '''Initialize an AIORedlock.

        Keyword arguments:
            key -- a string that identifies your resource
            masters -- the asyncio Redis clients used to achieve quorum for this
                AIORedlock's state
            raise_on_redis_errors -- whether to raise the QuorumIsImplssible
                exception when too many Redis masters throw errors
            auto_release_time -- the timeout in seconds by which to
                automatically release this AIORedlock, unless it's already been
                released
            num_extensions -- the number of times that this AIORedlock's lease
                can be extended
            context_manager_blocking -- when using this AIORedlock as a context
                manager, whether to block when acquiring
            context_manager_timeout -- if context_manager_blocking, how long to
                wait when acquiring before giving up and raising the
                QuorumNotAchieved exception
        '''
        if not context_manager_blocking and context_manager_timeout != -1:
            raise ValueError("can't specify a timeout for a non-blocking call")

        super().__init__(
            key=key,
            masters=masters,
            raise_on_redis_errors=raise_on_redis_errors,
        )
        self.auto_release_time = auto_release_time
        self.num_extensions = num_extensions
        self.context_manager_blocking = context_manager_blocking
        self.context_manager_timeout = context_manager_timeout
        self._uuid = ''
        self._extension_num = 0

    # Preserve the Open-Closed Principle with name mangling.
    #   https://youtu.be/miGolgp9xq8?t=2086
    #   https://stackoverflow.com/a/38534939
    async def __acquire_master(self, master: AIORedis) -> bool:  # type: ignore
        acquired = await master.set(
            self.key,
            self._uuid,
            px=int(self.auto_release_time * 1000),
            nx=True,
        )
        return bool(acquired)

    async def __acquired_master(self, master: AIORedis) -> int:  # type: ignore
        if self._uuid:
            ttl: int = await self._acquired_script(  # type: ignore
                keys=(self.key,),
                args=(self._uuid,),
                client=master,
            )
        else:
            ttl = 0
        return ttl

    async def __extend_master(self, master: AIORedis) -> bool:  # type: ignore
        auto_release_time_ms = int(self.auto_release_time * 1000)
        extended = await self._extend_script(  # type: ignore
            keys=(self.key,),
            args=(self._uuid, auto_release_time_ms),
            client=master,
        )
        return bool(extended)

    async def __release_master(self, master: AIORedis) -> bool:  # type: ignore
        released = await self._release_script(  # type: ignore
            keys=(self.key,),
            args=(self._uuid,),
            client=master,
        )
        return bool(released)

    def __drift(self) -> float:
        return self.auto_release_time * self._CLOCK_DRIFT_FACTOR + .002

    async def _acquire_masters(self,
                               *,
                               raise_on_redis_errors: bool | None = None,
                               ) -> bool:
        self._uuid = str(uuid.uuid4())
        self._extension_num = 0

        with ContextTimer() as timer:
            num_masters_acquired, redis_errors = 0, []
            coros = [self.__acquire_master(master) for master in self.masters]
            for coro in asyncio.as_completed(coros):
                try:
                    num_masters_acquired += await coro
                except RedisError as error:
                    redis_errors.append(error)
                    logger.exception(
                        '%s.__acquire_masters() caught %s',
                        self.__class__.__qualname__,
                        error.__class__.__qualname__,
                    )
            if num_masters_acquired > len(self.masters) // 2:
                validity_time = self.auto_release_time
                validity_time -= self.__drift()
                validity_time -= timer.elapsed() / 1000
                if validity_time > 0:
                    return True

        with contextlib.suppress(ReleaseUnlockedLock):
            await self.__release()
        self._check_enough_masters_up(raise_on_redis_errors, redis_errors)
        return False

    __acquire_masters = _acquire_masters

    async def acquire(self,
                      *,
                      blocking: bool = True,
                      timeout: float = -1,
                      raise_on_redis_errors: bool | None = None,
                      ) -> bool:
        '''Lock the lock.

        If blocking is True and timeout is -1, then wait for as long as
        necessary to acquire the lock.  Return True.

            >>> import asyncio
            >>> from redis.asyncio import Redis as AIORedis
            >>> async def main():
            ...     aioredis = AIORedis.from_url('redis://localhost:6379/1')
            ...     shower = AIORedlock(key='shower', masters={aioredis})
            ...     if await shower.acquire():
            ...         # Critical section - no other coroutine can enter while we hold the lock.
            ...         print(f"shower is {'occupied' if await shower.locked() else 'available'}")
            ...         await shower.release()
            ...     print(f"shower is {'occupied' if await shower.locked() else 'available'}")
            >>> asyncio.run(main(), debug=True)
            shower is occupied
            shower is available

        If blocking is True and timeout is not -1, then wait for up to timeout
        seconds to acquire the lock.  Return True if the lock was acquired;
        False if it wasn't.

            >>> async def main():
            ...     aioredis = AIORedis.from_url('redis://localhost:6379/1')
            ...     shower_lock_1 = AIORedlock(key='shower', masters={aioredis}, auto_release_time=.2)
            ...     shower_lock_2 = AIORedlock(key='shower', masters={aioredis}, auto_release_time=.2)
            ...     if await shower_lock_1.acquire():
            ...         print('shower_lock_1 acquired')
            ...     if await shower_lock_2.acquire(timeout=.5):
            ...         print('shower_lock_2 acquired')
            ...         await shower_lock_2.release()
            >>> asyncio.run(main(), debug=True)
            shower_lock_1 acquired
            shower_lock_2 acquired

            >>> async def main():
            ...     aioredis = AIORedis.from_url('redis://localhost:6379/1')
            ...     shower_lock_1 = AIORedlock(key='shower', masters={aioredis})
            ...     shower_lock_2 = AIORedlock(key='shower', masters={aioredis})
            ...     if await shower_lock_1.acquire():
            ...         print('shower_lock_1 acquired')
            ...     if not await shower_lock_2.acquire(timeout=.2):
            ...         print('shower_lock_2 not acquired')
            ...     await shower_lock_1.release()
            >>> asyncio.run(main(), debug=True)
            shower_lock_1 acquired
            shower_lock_2 not acquired

        If blocking is False and timeout is -1, then try just once right now to
        acquire the lock.  Return True if the lock was acquired; False if it
        wasn't.

            >>> async def main():
            ...     aioredis = AIORedis.from_url('redis://localhost:6379/1')
            ...     shower_lock_1 = AIORedlock(key='shower', masters={aioredis})
            ...     shower_lock_2 = AIORedlock(key='shower', masters={aioredis})
            ...     if await shower_lock_1.acquire():
            ...         print('shower_lock_1 acquired')
            ...     if not await shower_lock_2.acquire(blocking=False):
            ...         print('shower_lock_2 not acquired')
            ...     await shower_lock_1.release()
            >>> asyncio.run(main(), debug=True)
            shower_lock_1 acquired
            shower_lock_2 not acquired
        '''
        acquire_masters = functools.partial(
            self.__acquire_masters,
            raise_on_redis_errors=raise_on_redis_errors,
        )

        if blocking:
            enqueued = False
            with ContextTimer() as timer:
                while timeout == -1 or timer.elapsed() / 1000 < timeout:
                    if await acquire_masters():
                        if enqueued:
                            self.__log_time_enqueued(timer, acquired=True)
                        return True
                    enqueued = True
                    delay = random.uniform(0, self._RETRY_DELAY)  # nosec
                    await asyncio.sleep(delay)
            if enqueued:  # pragma: no cover
                self.__log_time_enqueued(timer, acquired=False)
            return False  # pragma: no cover

        if timeout == -1:
            return await acquire_masters()

        raise ValueError("can't specify a timeout for a non-blocking call")

    __acquire = acquire

    def __log_time_enqueued(self, timer: ContextTimer, *, acquired: bool) -> None:
        key_suffix = self.key.split(':', maxsplit=1)[1]
        time_enqueued = math.ceil(timer.elapsed())
        logger.info(
            'source=pottery sample#aioredlock.enqueued.%s=%dms sample#aioredlock.acquired.%s=%d',
            key_suffix,
            time_enqueued,
            key_suffix,
            acquired,
        )

    async def locked(self,
                     *,
                     raise_on_redis_errors: bool | None = None,
                     ) -> float:
        '''How much longer we'll hold the lock (unless we extend or release it).

        If we don't currently hold the lock, then this method returns 0.

            >>> async def main():
            ...     aioredis = AIORedis.from_url('redis://localhost:6379/1')
            ...     shower_lock = AIORedlock(key='shower', masters={aioredis})
            ...     print(await shower_lock.locked())
            ...     async with shower_lock:
            ...         print(math.ceil(await shower_lock.locked()))
            ...     print(await shower_lock.locked())
            >>> asyncio.run(main(), debug=True)
            0
            10
            0
        '''
        with ContextTimer() as timer:
            ttls, redis_errors = [], []
            coros = [self.__acquired_master(master) for master in self.masters]
            for coro in asyncio.as_completed(coros):
                try:
                    ttl = await coro / 1000
                except RedisError as error:
                    redis_errors.append(error)
                    logger.exception(
                        '%s.locked() caught %s',
                        self.__class__.__qualname__,
                        error.__class__.__qualname__,
                    )
                else:
                    if ttl:
                        ttls.append(ttl)
            if len(ttls) > len(self.masters) // 2:
                index = len(self.masters) // 2 - (not len(self.masters) % 2)
                validity_time: float = sorted(ttls)[index]
                validity_time -= self.__drift()
                validity_time -= timer.elapsed() / 1000
                return max(validity_time, 0)

        self._check_enough_masters_up(raise_on_redis_errors, redis_errors)
        return 0

    async def extend(self,
                     *,
                     raise_on_redis_errors: bool | None = None,
                     ) -> None:
        '''Extend our hold on the lock (if we currently hold it).

            >>> async def main():
            ...     aioredis = AIORedis.from_url('redis://localhost:6379/1')
            ...     shower_lock = AIORedlock(key='shower', masters={aioredis})
            ...     await shower_lock.acquire()
            ...     print(math.ceil(await shower_lock.locked()))
            ...     await asyncio.sleep(1)
            ...     print(math.ceil(await shower_lock.locked()))
            ...     await shower_lock.extend()
            ...     print(math.ceil(await shower_lock.locked()))
            ...     await shower_lock.release()
            >>> asyncio.run(main(), debug=True)
            10
            9
            10
        '''
        if self._extension_num >= self.num_extensions:
            raise TooManyExtensions(self.key, self.masters)

        num_masters_extended, redis_errors = 0, []
        coros = [self.__extend_master(master) for master in self.masters]
        for coro in asyncio.as_completed(coros):
            try:
                num_masters_extended += await coro
            except RedisError as error:
                redis_errors.append(error)
                logger.exception(
                    '%s.extend() caught %s',
                    self.__class__.__qualname__,
                    error.__class__.__qualname__,
                )
        if num_masters_extended > len(self.masters) // 2:
            self._extension_num += 1
            return

        self._check_enough_masters_up(raise_on_redis_errors, redis_errors)
        raise ExtendUnlockedLock(
            self.key,
            self.masters,
            redis_errors=redis_errors,
        )

    async def release(self,
                      *,
                      raise_on_redis_errors: bool | None = None,
                      ) -> None:
        '''Unlock the lock.

            >>> async def main():
            ...     aioredis = AIORedis.from_url('redis://localhost:6379/1')
            ...     shower_lock = AIORedlock(key='shower', masters={aioredis})
            ...     await shower_lock.acquire()
            ...     print(bool(await shower_lock.locked()))
            ...     await shower_lock.release()
            ...     print(bool(await shower_lock.locked()))
            >>> asyncio.run(main(), debug=True)
            True
            False
        '''
        num_masters_released, redis_errors = 0, []
        coros = [self.__release_master(master) for master in self.masters]
        for coro in asyncio.as_completed(coros):
            try:
                num_masters_released += await coro
            except RedisError as error:
                redis_errors.append(error)
                logger.exception(
                    '%s.release() caught %s',
                    self.__class__.__qualname__,
                    error.__class__.__qualname__,
                )
        if num_masters_released > len(self.masters) // 2:
            return

        self._check_enough_masters_up(raise_on_redis_errors, redis_errors)
        raise ReleaseUnlockedLock(
            self.key,
            self.masters,
            redis_errors=redis_errors,
        )

    __release = release

    async def __aenter__(self) -> AIORedlock:
        '''You can use an AIORedlock as a context manager.

            >>> async def main():
            ...     aioredis = AIORedis.from_url('redis://localhost:6379/1')
            ...     shower = AIORedlock(key='shower', masters={aioredis})
            ...     async with shower:
            ...         print(f"shower is {'occupied' if await shower.locked() else 'available'}")
            ...         # Critical section - no other coroutine can enter while we hold the lock.
            ...     print(f"shower is {'occupied' if await shower.locked() else 'available'}")
            >>> asyncio.run(main(), debug=True)
            shower is occupied
            shower is available
        '''
        acquired = await self.__acquire(
            blocking=self.context_manager_blocking,
            timeout=self.context_manager_timeout,
        )
        if acquired:
            return self
        raise QuorumNotAchieved(self.key, self.masters)

    async def __aexit__(self,
                        exc_type: Type[BaseException] | None,
                        exc_value: BaseException | None,
                        traceback: TracebackType | None,
                        ) -> None:
        '''You can use an AIORedlock as a context manager.

            >>> async def main():
            ...     aioredis = AIORedis.from_url('redis://localhost:6379/1')
            ...     shower = AIORedlock(key='shower', masters={aioredis})
            ...     async with shower:
            ...         print(f"shower is {'occupied' if await shower.locked() else 'available'}")
            ...         # Critical section - no other coroutine can enter while we hold the lock.
            ...     print(f"shower is {'occupied' if await shower.locked() else 'available'}")
            >>> asyncio.run(main(), debug=True)
            shower is occupied
            shower is available
        '''
        await self.__release()

    def __repr__(self) -> str:
        return f'<{self.__class__.__qualname__} key={self.key}>'
